@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2019 Taler Systems SA
+  (C) 2019, 2020, 2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -14,7 +14,7 @@
   TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */
 /**
- * @file backend/taler-merchant-httpd_config.c
+ * @file taler-merchant-httpd_config.c
  * @brief implement API for querying configuration data of the backend
  * @author Florian Dold
  */
@@ -25,8 +25,6 @@
 #include "taler-merchant-httpd.h"
 #include "taler-merchant-httpd_mhd.h"
 #include "taler-merchant-httpd_exchanges.h"
-#include "taler-merchant-httpd_tip-query.h"
-#include "taler-merchant-httpd_tip-reserve-helper.h"
 
 
 /**
@@ -41,41 +39,33 @@
  * versioning uses either.
  *
  * When changing this version, you likely want to also update
- * #TALER_PROTOCOL_CURRENT and #TALER_PROTOCOL_AGE in
- * TBD.c! // FIXME: update comment once libtalermerchant looks at version!
+ * #MERCHANT_PROTOCOL_CURRENT and #MERCHANT_PROTOCOL_AGE in
+ * merchant_api_config.c!
  */
-#define TALER_PROTOCOL_VERSION "0:0:0"
+#define MERCHANT_PROTOCOL_VERSION "1:0:0"
 
-
-/**
- * Handle a "/config" request.
- *
- * @param rh context of the handler
- * @param connection the MHD connection to handle
- * @param[in,out] connection_cls the connection's closure (can be updated)
- * @param upload_data upload data
- * @param[in,out] upload_data_size number of bytes (left) in @a upload_data
- * @param mi merchant backend instance, never NULL
- * @return MHD result code
- */
-int
+MHD_RESULT
 MH_handler_config (struct TMH_RequestHandler *rh,
                    struct MHD_Connection *connection,
-                   void **connection_cls,
-                   const char *upload_data,
-                   size_t *upload_data_size,
-                   struct MerchantInstance *mi)
+                   struct TMH_HandlerContext *hc)
 {
+  static struct MHD_Response *response;
+
   (void) rh;
-  (void) connection_cls;
-  (void) upload_data;
-  (void) upload_data_size;
-  (void) mi;
-  return TALER_MHD_reply_json_pack (connection,
-                                    MHD_HTTP_OK,
-                                    "{s:s, s:s}",
-                                    "currency", TMH_currency,
-                                    "version", TALER_PROTOCOL_VERSION);
+  (void) hc;
+  if (NULL == response)
+  {
+    response = TALER_MHD_MAKE_JSON_PACK (
+      GNUNET_JSON_pack_string ("currency",
+                               TMH_currency),
+      GNUNET_JSON_pack_string ("name",
+                               "taler-merchant"),
+      GNUNET_JSON_pack_string ("version",
+                               MERCHANT_PROTOCOL_VERSION));
+  }
+  return MHD_queue_response (connection,
+                             MHD_HTTP_OK,
+                             response);
 }
 
 
