@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2014-2020 Taler Systems SA
+  (C) 2014-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -119,16 +119,17 @@ struct GetOrderData
   enum TALER_ErrorCode ec;
 
   /**
-   * Did we suspend @a connection?
+   * Did we suspend @a connection and are thus in
+   * the #god_head DLL?
    */
   bool suspended;
 
   /**
-   * Set to true if we are dealing with an unclaimed order
-   * (and thus @e h_contract_terms is not set, and certain
+   * Set to true if we are dealing with a claimed order
+   * (and thus @e h_contract_terms is set, otherwise certain
    * DB queries will not work).
    */
-  bool unclaimed;
+  bool claimed;
 
   /**
    * Set to true if this payment has been refunded and
@@ -161,9 +162,6 @@ static struct GetOrderData *god_head;
 static struct GetOrderData *god_tail;
 
 
-/**
- * Force resuming all suspended order lookups, needed during shutdown.
- */
 void
 TMH_force_wallet_get_order_resume (void)
 {
@@ -192,6 +190,17 @@ suspend_god (struct GetOrderData *god)
   GNUNET_log (GNUNET_ERROR_TYPE_INFO,
               "Suspending GET /orders/%s\n",
               god->order_id);
+  if (NULL != god->contract_terms)
+  {
+    json_decref (god->contract_terms);
+    god->fulfillment_url = NULL;
+    god->contract_terms = NULL;
+  }
+  GNUNET_assert (! god->suspended);
+  god->suspended = true;
+  GNUNET_CONTAINER_DLL_insert (god_head,
+                               god_tail,
+                               god);
   TMH_long_poll_suspend (god->order_id,
                          god->session_id,
                          god->fulfillment_url,
@@ -231,7 +240,6 @@ make_taler_refund_uri (const char *merchant_base_url,
     return NULL;
   }
   GNUNET_assert (NULL != order_id);
-
   GNUNET_buffer_write_str (&buf,
                            "taler");
   if (0 == strcasecmp ("http",
@@ -275,23 +283,25 @@ TMH_make_order_status_url (struct MHD_Connection *con,
 
   host = MHD_lookup_connection_value (con,
                                       MHD_HEADER_KIND,
-                                      "Host");
+                                      MHD_HTTP_HEADER_HOST);
   forwarded_host = MHD_lookup_connection_value (con,
                                                 MHD_HEADER_KIND,
                                                 "X-Forwarded-Host");
-
   uri_path = MHD_lookup_connection_value (con,
                                           MHD_HEADER_KIND,
                                           "X-Forwarded-Prefix");
   if (NULL != forwarded_host)
     host = forwarded_host;
-
   if (NULL == host)
   {
     GNUNET_break (0);
     return NULL;
   }
-
+  if (NULL != strchr (host, '/'))
+  {
+    GNUNET_break_op (0);
+    return NULL;
+  }
   GNUNET_assert (NULL != instance_id);
   GNUNET_assert (NULL != order_id);
 
@@ -301,7 +311,6 @@ TMH_make_order_status_url (struct MHD_Connection *con,
   else
     GNUNET_buffer_write_str (&buf,
                              "https://");
-
   GNUNET_buffer_write_str (&buf,
                            host);
   if (NULL != uri_path)
@@ -322,6 +331,7 @@ TMH_make_order_status_url (struct MHD_Connection *con,
   if ((NULL != claim_token) &&
       (GNUNET_NO == GNUNET_is_zero (claim_token)))
   {
+    /* 'token=' for human readability */
     GNUNET_buffer_write_str (&buf,
                              "?token=");
     GNUNET_buffer_write_data_encoded (&buf,
@@ -374,26 +384,27 @@ TMH_make_taler_pay_uri (struct MHD_Connection *con,
 
   host = MHD_lookup_connection_value (con,
                                       MHD_HEADER_KIND,
-                                      "Host");
+                                      MHD_HTTP_HEADER_HOST);
   forwarded_host = MHD_lookup_connection_value (con,
                                                 MHD_HEADER_KIND,
                                                 "X-Forwarded-Host");
-
   uri_path = MHD_lookup_connection_value (con,
                                           MHD_HEADER_KIND,
                                           "X-Forwarded-Prefix");
   if (NULL != forwarded_host)
     host = forwarded_host;
-
   if (NULL == host)
   {
     GNUNET_break (0);
     return NULL;
   }
-
+  if (NULL != strchr (host, '/'))
+  {
+    GNUNET_break_op (0);
+    return NULL;
+  }
   GNUNET_assert (NULL != instance_id);
   GNUNET_assert (NULL != order_id);
-
   GNUNET_buffer_write_str (&buf,
                            "taler");
   if (GNUNET_NO == TALER_mhd_is_https (con))
@@ -421,6 +432,8 @@ TMH_make_taler_pay_uri (struct MHD_Connection *con,
   if ((NULL != claim_token) &&
       (GNUNET_NO == GNUNET_is_zero (claim_token)))
   {
+    /* Just 'c=' because this goes into QR
+       codes, so this is more compact. */
     GNUNET_buffer_write_str (&buf,
                              "?c=");
     GNUNET_buffer_write_data_encoded (&buf,
@@ -437,10 +450,10 @@ TMH_make_taler_pay_uri (struct MHD_Connection *con,
  * preferred language of the HTTP client.
  *
  * @param god order to extract summary from
- * @return NULL if no summary was provided in the contract
+ * @return dummy error message summary if no summary was provided in the contract
  */
 static const char *
-get_order_summary (struct GetOrderData *god)
+get_order_summary (const struct GetOrderData *god)
 {
   const char *language_pattern;
   const char *ret;
@@ -455,8 +468,10 @@ get_order_summary (struct GetOrderData *god)
                                                     "summary"));
   if (NULL == ret)
   {
-    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
-                "No order summary found!\n");
+    /* Upon order creation (and insertion into the database), the presence
+       of a summary should have been checked. So if we get here, someone
+       did something fishy to our database... */
+    GNUNET_break (0);
     ret = "<bug: no summary>";
   }
   return ret;
@@ -482,11 +497,12 @@ send_pay_request (struct GetOrderData *god,
   struct GNUNET_TIME_Relative remaining;
 
   remaining = GNUNET_TIME_absolute_get_remaining (god->sc.long_poll_timeout);
-  if ( (0 != remaining.rel_value_us) &&
-       ( (NULL == already_paid_order_id) ||
-         (NULL == god->fulfillment_url) ) )
+  if ( (! GNUNET_TIME_relative_is_zero (remaining)) &&
+       (NULL == already_paid_order_id) )
   {
     /* long polling: do not queue a response, suspend connection instead */
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Suspending request: long polling for payment\n");
     suspend_god (god);
     return MHD_YES;
   }
@@ -507,19 +523,28 @@ send_pay_request (struct GetOrderData *god,
                                                 god->hc->instance->settings.id,
                                                 &god->claim_token,
                                                 NULL);
+  if ( (NULL == taler_pay_uri) ||
+       (NULL == order_status_url) )
+  {
+    GNUNET_break_op (0);
+    GNUNET_free (taler_pay_uri);
+    GNUNET_free (order_status_url);
+    return TALER_MHD_reply_with_error (god->sc.con,
+                                       MHD_HTTP_BAD_REQUEST,
+                                       TALER_EC_GENERIC_HTTP_HEADERS_MALFORMED,
+                                       "host");
+  }
   if (god->generate_html)
   {
-    char *qr;
-
-    if ( (NULL != already_paid_order_id) &&
-         (NULL != god->fulfillment_url) )
+    if (NULL != already_paid_order_id)
     {
       struct MHD_Response *reply;
-      MHD_RESULT ret;
 
+      GNUNET_assert (NULL != god->fulfillment_url);
       GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
-                  "Redirecting to already paid order %s\n",
-                  already_paid_order_id);
+                  "Redirecting to already paid order %s via fulfillment URL %s\n",
+                  already_paid_order_id,
+                  god->fulfillment_url);
       reply = MHD_create_response_from_buffer (0,
                                                NULL,
                                                MHD_RESPMEM_PERSISTENT);
@@ -532,60 +557,72 @@ send_pay_request (struct GetOrderData *god,
                     MHD_add_response_header (reply,
                                              MHD_HTTP_HEADER_LOCATION,
                                              god->fulfillment_url));
-      ret = MHD_queue_response (god->sc.con,
-                                MHD_HTTP_FOUND,
-                                reply);
-      MHD_destroy_response (reply);
-      return ret;
+      {
+        MHD_RESULT ret;
+
+        ret = MHD_queue_response (god->sc.con,
+                                  MHD_HTTP_FOUND,
+                                  reply);
+        MHD_destroy_response (reply);
+        return ret;
+      }
     }
 
-    qr = TMH_create_qrcode (taler_pay_uri);
-    if (NULL == qr)
     {
-      GNUNET_break (0);
-      return MHD_NO;
-    }
-    {
-      enum GNUNET_GenericReturnValue res;
-      json_t *context;
+      char *qr;
 
-      context = json_pack ("{s:s, s:s, s:s, s:s}",
-                           "taler_pay_uri",
-                           taler_pay_uri,
-                           "order_status_url",
-                           order_status_url,
-                           "taler_pay_qrcode_svg",
-                           qr,
-                           "order_summary",
-                           get_order_summary (god));
-      GNUNET_assert (NULL != context);
-      res = TMH_return_from_template (god->sc.con,
-                                      MHD_HTTP_PAYMENT_REQUIRED,
-                                      "request_payment",
-                                      taler_pay_uri,
-                                      context);
-      if (GNUNET_SYSERR == res)
+      qr = TMH_create_qrcode (taler_pay_uri);
+      if (NULL == qr)
       {
         GNUNET_break (0);
-        ret = MHD_NO;
+        return MHD_NO;
       }
-      ret = MHD_YES;
-      json_decref (context);
-    }
-    GNUNET_free (qr);
-  }
-  else
-  {
-    ret = TALER_MHD_reply_json_pack (god->sc.con,
-                                     MHD_HTTP_PAYMENT_REQUIRED,
-                                     "{s:s, s:s?, s:s?}",
-                                     "taler_pay_uri",
-                                     taler_pay_uri,
-                                     "fulfillment_url",
-                                     god->fulfillment_url,
-                                     "already_paid_order_id",
-                                     already_paid_order_id);
+      {
+        enum GNUNET_GenericReturnValue res;
+        json_t *context;
 
+        context = GNUNET_JSON_PACK (
+          GNUNET_JSON_pack_string ("taler_pay_uri",
+                                   taler_pay_uri),
+          GNUNET_JSON_pack_string ("order_status_url",
+                                   order_status_url),
+          GNUNET_JSON_pack_string ("taler_pay_qrcode_svg",
+                                   qr),
+          GNUNET_JSON_pack_string ("order_summary",
+                                   get_order_summary (god)));
+        res = TMH_return_from_template (god->sc.con,
+                                        MHD_HTTP_PAYMENT_REQUIRED,
+                                        "request_payment",
+                                        god->hc->instance->settings.id,
+                                        taler_pay_uri,
+                                        context);
+        if (GNUNET_SYSERR == res)
+        {
+          GNUNET_break (0);
+          ret = MHD_NO;
+        }
+        else
+        {
+          ret = MHD_YES;
+        }
+        json_decref (context);
+      }
+      GNUNET_free (qr);
+    }
+  }
+  else /* end of 'generate HTML' */
+  {
+    ret = TALER_MHD_REPLY_JSON_PACK (
+      god->sc.con,
+      MHD_HTTP_PAYMENT_REQUIRED,
+      GNUNET_JSON_pack_string ("taler_pay_uri",
+                               taler_pay_uri),
+      GNUNET_JSON_pack_allow_null (
+        GNUNET_JSON_pack_string ("fulfillment_url",
+                                 god->fulfillment_url)),
+      GNUNET_JSON_pack_allow_null (
+        GNUNET_JSON_pack_string ("already_paid_order_id",
+                                 already_paid_order_id)));
   }
   GNUNET_free (taler_pay_uri);
   GNUNET_free (order_status_url);
@@ -625,6 +662,7 @@ process_refunds_cb (void *cls,
               TALER_amount2s (refund_amount),
               TALER_B2S (coin_pub),
               reason);
+  god->refund_available |= pending;
   if (god->refunded)
   {
     GNUNET_assert (0 <=
@@ -635,7 +673,6 @@ process_refunds_cb (void *cls,
   }
   god->refund_amount = *refund_amount;
   god->refunded = true;
-  god->refund_available |= pending;
 }
 
 
@@ -650,7 +687,10 @@ god_cleanup (void *cls)
   struct GetOrderData *god = cls;
 
   if (NULL != god->contract_terms)
+  {
     json_decref (god->contract_terms);
+    god->contract_terms = NULL;
+  }
   GNUNET_free (god);
 }
 
@@ -665,6 +705,8 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
   enum GNUNET_DB_QueryStatus qs;
   bool contract_match = false;
   bool token_match = false;
+  bool h_contract_provided = false;
+  bool claim_token_provided = false;
   bool contract_available = false;
   const char *merchant_base_url;
 
@@ -676,6 +718,31 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
     god->sc.con = connection;
     god->hc = hc;
     god->order_id = order_id;
+    god->generate_html = TMH_MHD_test_html_desired (connection);
+
+
+    /* first-time initialization / sanity checks */
+    {
+      const char *cts;
+
+      cts = MHD_lookup_connection_value (connection,
+                                         MHD_GET_ARGUMENT_KIND,
+                                         "h_contract");
+      if ( (NULL != cts) &&
+           (GNUNET_OK !=
+            GNUNET_CRYPTO_hash_from_string (cts,
+                                            &god->h_contract_terms)) )
+      {
+        /* cts has wrong encoding */
+        GNUNET_break_op (0);
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_BAD_REQUEST,
+                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
+                                           "h_contract");
+      }
+      if (NULL != cts)
+        h_contract_provided = true;
+    }
 
     {
       const char *ct;
@@ -697,43 +764,27 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
                                            TALER_EC_GENERIC_PARAMETER_MALFORMED,
                                            "token");
       }
+      if (NULL != ct)
+        claim_token_provided = true;
     }
-
-    {
-      const char *cts;
-
-      cts = MHD_lookup_connection_value (connection,
-                                         MHD_GET_ARGUMENT_KIND,
-                                         "h_contract");
-      if ( (NULL != cts) &&
-           (GNUNET_OK !=
-            GNUNET_CRYPTO_hash_from_string (cts,
-                                            &god->h_contract_terms)) )
-      {
-        /* cts has wrong encoding */
-        GNUNET_break_op (0);
-        return TALER_MHD_reply_with_error (connection,
-                                           MHD_HTTP_BAD_REQUEST,
-                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
-                                           "h_contract");
-      }
-    }
-
-    god->generate_html = TMH_MHD_test_html_desired (connection);
+    god->session_id = MHD_lookup_connection_value (connection,
+                                                   MHD_GET_ARGUMENT_KIND,
+                                                   "session_id");
     {
       const char *long_poll_timeout_ms;
 
       long_poll_timeout_ms = MHD_lookup_connection_value (connection,
                                                           MHD_GET_ARGUMENT_KIND,
                                                           "timeout_ms");
-      if ((NULL != long_poll_timeout_ms) &&
-          ! god->generate_html)
+      if (NULL != long_poll_timeout_ms)
       {
         unsigned int timeout;
+        char dummy;
 
         if (1 != sscanf (long_poll_timeout_ms,
-                         "%u",
-                         &timeout))
+                         "%u%c",
+                         &timeout,
+                         &dummy))
         {
           GNUNET_break_op (0);
           return TALER_MHD_reply_with_error (connection,
@@ -741,14 +792,12 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
                                              TALER_EC_GENERIC_PARAMETER_MALFORMED,
                                              "timeout_ms (must be non-negative number)");
         }
-        god->sc.long_poll_timeout
-          = GNUNET_TIME_relative_to_absolute (GNUNET_TIME_relative_multiply (
-                                                GNUNET_TIME_UNIT_MILLISECONDS,
-                                                timeout));
-      }
-      else
-      {
-        god->sc.long_poll_timeout = GNUNET_TIME_UNIT_ZERO_ABS;
+        /* If HTML is requested, we never long poll. Makes no sense */
+        if (! god->generate_html)
+          god->sc.long_poll_timeout
+            = GNUNET_TIME_relative_to_absolute (GNUNET_TIME_relative_multiply (
+                                                  GNUNET_TIME_UNIT_MILLISECONDS,
+                                                  timeout));
       }
     }
 
@@ -759,7 +808,6 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
         MHD_lookup_connection_value (connection,
                                      MHD_GET_ARGUMENT_KIND,
                                      "await_refund_obtained");
-
       god->sc.awaiting_refund_obtained =
         (NULL != await_refund_obtained_s)
         ? 0 == strcasecmp (await_refund_obtained_s, "yes")
@@ -789,38 +837,50 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
         god->sc.awaiting_refund = true;
       }
     }
-
-    god->session_id = MHD_lookup_connection_value (connection,
-                                                   MHD_GET_ARGUMENT_KIND,
-                                                   "session_id");
   } /* end of first-time initialization / sanity checks */
 
+  if (god->suspended)
+  {
+    god->suspended = false;
+    GNUNET_CONTAINER_DLL_remove (god_head,
+                                 god_tail,
+                                 god);
+  }
+
   /* Convert order_id to h_contract_terms */
   TMH_db->preflight (TMH_db->cls);
   if (NULL == god->contract_terms)
   {
     uint64_t order_serial;
+    struct TALER_ClaimTokenP db_claim_token;
 
     qs = TMH_db->lookup_contract_terms (TMH_db->cls,
                                         hc->instance->settings.id,
                                         order_id,
                                         &god->contract_terms,
-                                        &order_serial);
+                                        &order_serial,
+                                        &db_claim_token);
     if (0 > qs)
     {
       /* single, read-only SQL statements should never cause
          serialization problems */
       GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
       /* Always report on hard error as well to enable diagnostics */
-      GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
+      GNUNET_break (0);
       return TALER_MHD_reply_with_error (connection,
                                          MHD_HTTP_INTERNAL_SERVER_ERROR,
                                          TALER_EC_GENERIC_DB_FETCH_FAILED,
-                                         "contract terms");
+                                         "lookup_contract_terms");
     }
+
+    /* Note: when "!ord.requireClaimToken" and the client does not provide
+       a claim token (all zeros!), then token_match==TRUE below: */
+    token_match = (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT == qs)
+                  && (0 == GNUNET_memcmp (&db_claim_token,
+                                          &god->claim_token));
   }
 
-  /* Check client provided the right hash code of the contract terms */
+  /* Check if client provided the right hash code of the contract terms */
   if (NULL != god->contract_terms)
   {
     struct GNUNET_HashCode h;
@@ -831,7 +891,6 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
                                   &h))
     {
       GNUNET_break (0);
-      GNUNET_free (god);
       return TALER_MHD_reply_with_error (connection,
                                          MHD_HTTP_INTERNAL_SERVER_ERROR,
                                          TALER_EC_GENERIC_FAILED_COMPUTE_JSON_HASH,
@@ -853,7 +912,11 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
     }
   }
 
-  if (! contract_match)
+  if (contract_available)
+  {
+    god->claimed = true;
+  }
+  else
   {
     struct TALER_ClaimTokenP db_claim_token;
     struct GNUNET_HashCode unused;
@@ -872,16 +935,13 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
          serialization problems */
       GNUNET_break (GNUNET_DB_STATUS_SOFT_ERROR != qs);
       /* Always report on hard error as well to enable diagnostics */
-      GNUNET_break (GNUNET_DB_STATUS_HARD_ERROR == qs);
+      GNUNET_break (0);
       return TALER_MHD_reply_with_error (connection,
                                          MHD_HTTP_INTERNAL_SERVER_ERROR,
                                          TALER_EC_GENERIC_DB_FETCH_FAILED,
-                                         "order");
+                                         "lookup_order");
     }
-    god->unclaimed = (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT == qs) &&
-                     ! contract_available;
-    if ( (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs) &&
-         (NULL == god->contract_terms) )
+    if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
     {
       GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                   "Unknown order id given: `%s'\n",
@@ -891,10 +951,14 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
                                          TALER_EC_MERCHANT_GENERIC_ORDER_UNKNOWN,
                                          order_id);
     }
-    token_match = (0 == GNUNET_memcmp (&db_claim_token,
+    /* Note: when "!ord.requireClaimToken" and the client does not provide
+       a claim token (all zeros!), then token_match==TRUE below: */
+    token_match = (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT == qs) &&
+                  (0 == GNUNET_memcmp (&db_claim_token,
                                        &god->claim_token));
-  }   /* end unclaimed order logic */
+  } /* end unclaimed order logic */
 
+  GNUNET_assert (NULL != god->contract_terms);
   merchant_base_url = json_string_value (json_object_get (god->contract_terms,
                                                           "merchant_base_url"));
   if (NULL == merchant_base_url)
@@ -911,38 +975,63 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
                                                 god->contract_terms,
                                                 "fulfillment_url"));
   GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
-              "Token match: %d, contract match: %d, unclaimed: %d\n",
+              "Token match: %d, contract_available: %d, contract match: %d, claimed: %d\n",
               token_match,
+              contract_available,
               contract_match,
-              god->unclaimed);
-  if ( (god->unclaimed) &&
-       (! token_match) )
+              god->claimed);
+
+  if (claim_token_provided && ! token_match)
   {
-    /* Token wrong, and required because contract is unclaimed */
+    /* Authentication provided but wrong. */
     GNUNET_break_op (0);
     return TALER_MHD_reply_with_error (connection,
                                        MHD_HTTP_FORBIDDEN,
                                        TALER_EC_MERCHANT_GET_ORDERS_ID_INVALID_TOKEN,
-                                       NULL);
+                                       "authentication with claim token provided but wrong");
   }
-  if ( ( (! token_match) ||
-         (GNUNET_YES == GNUNET_is_zero (&god->claim_token)) ) &&
-       (! contract_match) )
+
+  if (h_contract_provided && ! contract_match)
   {
-    if (NULL == god->fulfillment_url)
+    /* Authentication provided but wrong. */
+    GNUNET_break_op (0);
+    /* FIXME: use better error code */
+    return TALER_MHD_reply_with_error (connection,
+                                       MHD_HTTP_FORBIDDEN,
+                                       TALER_EC_MERCHANT_GET_ORDERS_ID_INVALID_TOKEN,
+                                       "authentication with h_contract provided but wrong");
+  }
+
+  if (! (token_match ||
+         contract_match) )
+  {
+    const char *public_reorder_url;
+
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Neither claim token nor contract matched\n");
+    public_reorder_url = json_string_value (json_object_get (
+                                              god->contract_terms,
+                                              "public_reorder_url"));
+    /* Client has no rights to this order */
+    if (NULL == public_reorder_url)
     {
-      if (GNUNET_NO ==
-          GNUNET_is_zero (&god->h_contract_terms))
+      /* We cannot give the client a new order, just fail */
+      if (! GNUNET_is_zero (&god->h_contract_terms))
+      {
+        GNUNET_break_op (0);
         return TALER_MHD_reply_with_error (
           connection,
           MHD_HTTP_FORBIDDEN,
           TALER_EC_MERCHANT_GENERIC_CONTRACT_HASH_DOES_NOT_MATCH_ORDER,
           NULL);
+      }
       return TALER_MHD_reply_with_error (connection,
                                          MHD_HTTP_FORBIDDEN,
                                          TALER_EC_MERCHANT_GET_ORDERS_ID_INVALID_TOKEN,
-                                         NULL);
+                                         "no 'public_reorder_url'");
     }
+    /* We have a fulfillment URL, redirect the client there, maybe
+       the frontend can generate a fresh order for this new customer */
     if (god->generate_html)
     {
       /* Contract was claimed (maybe by another device), so this client
@@ -966,7 +1055,7 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
       GNUNET_break (MHD_YES ==
                     MHD_add_response_header (reply,
                                              MHD_HTTP_HEADER_LOCATION,
-                                             god->fulfillment_url));
+                                             public_reorder_url));
       ret = MHD_queue_response (connection,
                                 MHD_HTTP_FOUND,
                                 reply);
@@ -974,28 +1063,27 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
       return ret;
     }
     /* Need to generate JSON reply */
-    return TALER_MHD_reply_json_pack (connection,
-                                      MHD_HTTP_ACCEPTED,
-                                      "{s:s}",
-                                      "fulfillment_url",
-                                      god->fulfillment_url);
+    return TALER_MHD_REPLY_JSON_PACK (
+      connection,
+      MHD_HTTP_ACCEPTED,
+      GNUNET_JSON_pack_string ("public_reorder_url",
+                               public_reorder_url));
   }
 
-  if (god->unclaimed)
-  {
-    /* Order is unclaimed, no need to check for payments or even
-       refunds, simply always generate payment request */
-    GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
-                "Order unclaimed, sending pay request for order %s\n",
-                order_id);
-    return send_pay_request (god,
-                             NULL);
-  }
+  GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+              "Claim token or contract matched\n");
 
   if ( (NULL != god->session_id) &&
        (NULL != god->fulfillment_url) )
   {
-    /* Check if paid within a session. */
+    /* Check if client paid for this fulfillment article
+       already within this session, but using a different
+       order ID. If so, redirect the client to the order
+       it already paid.  Allows, for example, the case
+       where a mobile phone pays for a browser's session,
+       where the mobile phone has a different order
+       ID (because it purchased the article earlier)
+       than the one that the browser is waiting for. */
     char *already_paid_order_id = NULL;
 
     GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
@@ -1025,7 +1113,7 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
     {
       MHD_RESULT ret;
 
-      GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                   "Sending pay request for order %s (already paid: %s)\n",
                   order_id,
                   already_paid_order_id);
@@ -1034,10 +1122,21 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
       GNUNET_free (already_paid_order_id);
       return ret;
     }
-    GNUNET_break (1 == qs);
+    GNUNET_break (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT == qs);
     GNUNET_free (already_paid_order_id);
   }
 
+  if (! god->claimed)
+  {
+    /* Order is unclaimed, no need to check for payments or even
+       refunds, simply always generate payment request */
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Order unclaimed, sending pay request for order %s\n",
+                order_id);
+    return send_pay_request (god,
+                             NULL);
+  }
+
   {
     /* Check if paid. */
     struct GNUNET_HashCode h_contract;
@@ -1055,15 +1154,12 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
       return TALER_MHD_reply_with_error (connection,
                                          MHD_HTTP_INTERNAL_SERVER_ERROR,
                                          TALER_EC_GENERIC_DB_FETCH_FAILED,
-                                         "order status");
+                                         "lookup_order_status");
     }
     GNUNET_break (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT == qs);
-    GNUNET_break (0 ==
-                  GNUNET_memcmp (&h_contract,
-                                 &god->h_contract_terms));
     if (! paid)
     {
-      GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                   "Order claimed but unpaid, sending pay request for order %s\n",
                   order_id);
       return send_pay_request (god,
@@ -1074,7 +1170,7 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
   /* At this point, we know the contract was paid. Let's check for
      refunds. First, clear away refunds found from previous invocations. */
   GNUNET_assert (GNUNET_OK ==
-                 TALER_amount_get_zero (TMH_currency,
+                 TALER_amount_set_zero (TMH_currency,
                                         &god->refund_amount));
   qs = TMH_db->lookup_refunds_detailed (TMH_db->cls,
                                         hc->instance->settings.id,
@@ -1087,7 +1183,7 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
     return TALER_MHD_reply_with_error (connection,
                                        MHD_HTTP_INTERNAL_SERVER_ERROR,
                                        TALER_EC_GENERIC_DB_FETCH_FAILED,
-                                       "refunds, detailed");
+                                       "lookup_refunds_detailed");
   }
 
   if ( ((god->sc.awaiting_refund) &&
@@ -1102,7 +1198,7 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
     struct GNUNET_TIME_Relative remaining;
 
     remaining = GNUNET_TIME_absolute_get_remaining (god->sc.long_poll_timeout);
-    if (0 != remaining.rel_value_us)
+    if (! GNUNET_TIME_relative_is_zero (remaining))
     {
       /* yes, indeed suspend */
       if (god->sc.awaiting_refund)
@@ -1110,7 +1206,7 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
                     "Awaiting refund exceeding %s\n",
                     TALER_amount2s (&god->sc.refund_expected));
       if (god->sc.awaiting_refund_obtained)
-        GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+        GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                     "Awaiting pending refunds\n");
       suspend_god (god);
       return MHD_YES;
@@ -1118,97 +1214,97 @@ TMH_get_orders_ID (const struct TMH_RequestHandler *rh,
   }
 
   /* All operations done, build final response */
+  if (god->generate_html)
   {
-    if (god->generate_html)
+    enum GNUNET_GenericReturnValue res;
+
+    if (god->refund_available)
     {
-      enum GNUNET_GenericReturnValue res;
+      char *qr;
+      char *uri;
 
-      if (god->refund_available)
+      GNUNET_assert (NULL != god->contract_terms);
+      uri = make_taler_refund_uri (merchant_base_url,
+                                   order_id);
+      if (NULL == uri)
       {
-        char *qr;
-        char *uri;
-
-        GNUNET_assert (NULL != god->contract_terms);
-        uri = make_taler_refund_uri (merchant_base_url,
-                                     order_id);
-        if (NULL == uri)
-        {
-          GNUNET_break (0);
-          return TALER_MHD_reply_with_error (god->sc.con,
-                                             MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                             TALER_EC_GENERIC_ALLOCATION_FAILURE,
-                                             "refund URI");
-        }
-        qr = TMH_create_qrcode (uri);
-        if (NULL == qr)
-        {
-          GNUNET_break (0);
-          GNUNET_free (uri);
-          return TALER_MHD_reply_with_error (god->sc.con,
-                                             MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                             TALER_EC_GENERIC_ALLOCATION_FAILURE,
-                                             "qr code");
-        }
-        {
-          json_t *context;
-
-          context = json_pack ("{s:s, s:s, s:s, s:s}",
-                               "order_summary",
-                               get_order_summary (god),
-                               "refund_amount",
-                               TALER_amount2s (&god->refund_amount),
-                               "taler_refund_uri",
-                               uri,
-                               "taler_refund_qrcode_svg",
-                               qr);
-          GNUNET_assert (NULL != context);
-          res = TMH_return_from_template (god->sc.con,
-                                          MHD_HTTP_OK,
-                                          "offer_refund",
-                                          uri,
-                                          context);
-          json_decref (context);
-        }
+        GNUNET_break (0);
+        return TALER_MHD_reply_with_error (god->sc.con,
+                                           MHD_HTTP_INTERNAL_SERVER_ERROR,
+                                           TALER_EC_GENERIC_ALLOCATION_FAILURE,
+                                           "refund URI");
+      }
+      qr = TMH_create_qrcode (uri);
+      if (NULL == qr)
+      {
+        GNUNET_break (0);
         GNUNET_free (uri);
-        GNUNET_free (qr);
+        return TALER_MHD_reply_with_error (god->sc.con,
+                                           MHD_HTTP_INTERNAL_SERVER_ERROR,
+                                           TALER_EC_GENERIC_ALLOCATION_FAILURE,
+                                           "qr code");
       }
-      else
       {
         json_t *context;
 
-        context = json_pack ("{s:O, s:s, s:s}",
-                             "contract_terms",
-                             god->contract_terms,
-                             "order_summary",
-                             get_order_summary (god),
-                             "refund_amount",
-                             TALER_amount2s (&god->refund_amount));
-        GNUNET_assert (NULL != context);
+        context = GNUNET_JSON_PACK (
+          GNUNET_JSON_pack_string ("order_summary",
+                                   get_order_summary (god)),
+          TALER_JSON_pack_amount ("refund_amount",
+                                  &god->refund_amount),
+          GNUNET_JSON_pack_string ("taler_refund_uri",
+                                   uri),
+          GNUNET_JSON_pack_string ("taler_refund_qrcode_svg",
+                                   qr));
         res = TMH_return_from_template (god->sc.con,
                                         MHD_HTTP_OK,
-                                        "show_order_details",
-                                        NULL,
+                                        "offer_refund",
+                                        hc->instance->settings.id,
+                                        uri,
                                         context);
         json_decref (context);
       }
-      if (GNUNET_SYSERR == res)
-      {
-        GNUNET_break (0);
-        return MHD_NO;
-      }
-      return MHD_YES;
+      GNUNET_free (uri);
+      GNUNET_free (qr);
     }
     else
     {
-      return TALER_MHD_reply_json_pack (
-        connection,
-        MHD_HTTP_OK,
-        "{s:b, s:b, s:o}",
-        "refunded", god->refunded,
-        "refund_pending", god->refund_available,
-        "refund_amount", TALER_JSON_from_amount (&god->refund_amount));
+      json_t *context;
+
+      context = GNUNET_JSON_PACK (
+        GNUNET_JSON_pack_object_incref ("contract_terms",
+                                        god->contract_terms),
+        GNUNET_JSON_pack_string ("order_summary",
+                                 get_order_summary (god)),
+        TALER_JSON_pack_amount ("refund_amount",
+                                &god->refund_amount));
+      res = TMH_return_from_template (god->sc.con,
+                                      MHD_HTTP_OK,
+                                      "show_order_details",
+                                      hc->instance->settings.id,
+                                      NULL,
+                                      context);
+      json_decref (context);
     }
+    if (GNUNET_SYSERR == res)
+    {
+      GNUNET_break (0);
+      return MHD_NO;
+    }
+    return MHD_YES;
   }
+  return TALER_MHD_REPLY_JSON_PACK (
+    connection,
+    MHD_HTTP_OK,
+    GNUNET_JSON_pack_allow_null (
+      GNUNET_JSON_pack_string ("fulfillment_url",
+                               god->fulfillment_url)),
+    GNUNET_JSON_pack_bool ("refunded",
+                           god->refunded),
+    GNUNET_JSON_pack_bool ("refund_pending",
+                           god->refund_available),
+    TALER_JSON_pack_amount ("refund_amount",
+                            &god->refund_amount));
 }
 
 
