@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2017-2020 Taler Systems SA
+  (C) 2017-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
@@ -25,6 +25,7 @@
 #include <taler/taler_signatures.h>
 #include "taler-merchant-httpd.h"
 #include "taler-merchant-httpd_mhd.h"
+#include "taler-merchant-httpd_helper.h"
 #include "taler-merchant-httpd_exchanges.h"
 #include "taler-merchant-httpd_post-tips-ID-pickup.h"
 
@@ -223,6 +224,7 @@ stop_operations (struct PickupContext *pc)
     GNUNET_CONTAINER_DLL_remove (pc->po_head,
                                  pc->po_tail,
                                  po);
+    GNUNET_free (po);
   }
 }
 
@@ -287,15 +289,13 @@ withdraw_cb (void *cls,
                                po);
   if (NULL == blind_sig)
   {
+    GNUNET_free (po);
     stop_operations (pc);
-    pc->http_status = MHD_HTTP_FAILED_DEPENDENCY;
+    pc->http_status = MHD_HTTP_BAD_GATEWAY;
     pc->response =
-      TALER_MHD_make_json_pack (
-        "{s:I, s:I, s:I, s:O}",
-        "code", (json_int_t) TALER_EC_MERCHANT_TIP_PICKUP_EXCHANGE_ERROR,
-        "exchange_code", (json_int_t) hr->ec,
-        "exchange_http_status", (json_int_t) hr->http_status,
-        "exchange_reply", hr->reply);
+      TALER_MHD_MAKE_JSON_PACK (
+        TALER_JSON_pack_ec (TALER_EC_MERCHANT_TIP_PICKUP_EXCHANGE_ERROR),
+        TMH_pack_exchange_reply (hr));
     MHD_resume_connection (pc->connection);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
     return;
@@ -304,6 +304,7 @@ withdraw_cb (void *cls,
                                               &pc->pickup_id,
                                               po->offset,
                                               blind_sig);
+  GNUNET_free (po);
   if (qs < 0)
   {
     stop_operations (pc);
@@ -354,14 +355,11 @@ do_withdraw (void *cls,
     GNUNET_CONTAINER_DLL_remove (pc->po_head,
                                  pc->po_tail,
                                  po);
+    GNUNET_free (po);
     pc->http_status = MHD_HTTP_GATEWAY_TIMEOUT;
-    pc->response = TALER_MHD_make_json_pack (
-      "{s:s, s:I}",
-      "hint",
-      TALER_ErrorCode_get_hint (
-        TALER_EC_MERCHANT_GENERIC_EXCHANGE_TIMEOUT),
-      "code",
-      (json_int_t) TALER_EC_MERCHANT_GENERIC_EXCHANGE_TIMEOUT);
+    pc->response = TALER_MHD_MAKE_JSON_PACK (
+      TALER_JSON_pack_ec (
+        TALER_EC_MERCHANT_GENERIC_EXCHANGE_TIMEOUT));
     MHD_resume_connection (pc->connection);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
     return;
@@ -372,16 +370,13 @@ do_withdraw (void *cls,
     GNUNET_CONTAINER_DLL_remove (pc->po_head,
                                  pc->po_tail,
                                  po);
+    GNUNET_free (po);
     pc->http_status = MHD_HTTP_BAD_GATEWAY;
     pc->response =
-      TALER_MHD_make_json_pack (
-        "{s:s, s:I, s:I, s:I, s:O?}",
-        "hint", TALER_ErrorCode_get_hint (
+      TALER_MHD_MAKE_JSON_PACK (
+        TALER_JSON_pack_ec (
           TALER_EC_MERCHANT_GENERIC_EXCHANGE_CONNECT_FAILURE),
-        "code", (json_int_t) TALER_EC_MERCHANT_GENERIC_EXCHANGE_CONNECT_FAILURE,
-        "exchange_code", (json_int_t) hr->ec,
-        "exchange_http_status", (json_int_t) hr->http_status,
-        "exchange_reply", hr->reply);
+        TMH_pack_exchange_reply (hr));
     MHD_resume_connection (pc->connection);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
     return;
@@ -480,10 +475,8 @@ compute_total_requested (void *cls,
   if (NULL == hr)
   {
     pc->http_status = MHD_HTTP_GATEWAY_TIMEOUT;
-    pc->response = TALER_MHD_make_json_pack (
-      "{s:I, s:s}",
-      "code", (json_int_t) TALER_EC_MERCHANT_GENERIC_EXCHANGE_TIMEOUT,
-      "hint", TALER_ErrorCode_get_hint (
+    pc->response = TALER_MHD_MAKE_JSON_PACK (
+      TALER_JSON_pack_ec (
         TALER_EC_MERCHANT_GENERIC_EXCHANGE_TIMEOUT));
     MHD_resume_connection (pc->connection);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
@@ -493,14 +486,10 @@ compute_total_requested (void *cls,
   {
     pc->http_status = MHD_HTTP_BAD_GATEWAY;
     pc->response =
-      TALER_MHD_make_json_pack (
-        "{s:I, s:s, s:I, s:I, s:O}",
-        "code", (json_int_t) TALER_EC_MERCHANT_GENERIC_EXCHANGE_CONNECT_FAILURE,
-        "hint", TALER_ErrorCode_get_hint (
+      TALER_MHD_MAKE_JSON_PACK (
+        TALER_JSON_pack_ec (
           TALER_EC_MERCHANT_GENERIC_EXCHANGE_CONNECT_FAILURE),
-        "exchange_code", (json_int_t) hr->ec,
-        "exchange_http_status", (json_int_t) hr->http_status,
-        "exchange_reply", hr->reply);
+        TMH_pack_exchange_reply (hr));
     MHD_resume_connection (pc->connection);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
     return;
@@ -509,20 +498,16 @@ compute_total_requested (void *cls,
   {
     pc->http_status = MHD_HTTP_BAD_GATEWAY;
     pc->response =
-      TALER_MHD_make_json_pack (
-        "{s:I, s:s, s:I, s:I, s:O}",
-        "code", (json_int_t) TALER_EC_MERCHANT_GENERIC_EXCHANGE_KEYS_FAILURE,
-        "hint", TALER_ErrorCode_get_hint (
+      TALER_MHD_MAKE_JSON_PACK (
+        TALER_JSON_pack_ec (
           TALER_EC_MERCHANT_GENERIC_EXCHANGE_KEYS_FAILURE),
-        "exchange_code", (json_int_t) hr->ec,
-        "exchange_http_status", (json_int_t) hr->http_status,
-        "exchange_reply", hr->reply);
+        TMH_pack_exchange_reply (hr));
     MHD_resume_connection (pc->connection);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
     return;
   }
   GNUNET_assert (GNUNET_OK ==
-                 TALER_amount_get_zero (TMH_currency,
+                 TALER_amount_set_zero (TMH_currency,
                                         &pc->total_requested));
   for (unsigned int i = 0; i<pc->planchets_length; i++)
   {
@@ -535,13 +520,10 @@ compute_total_requested (void *cls,
     {
       pc->http_status = MHD_HTTP_CONFLICT;
       pc->response =
-        TALER_MHD_make_json_pack (
-          "{s:I, s:I, s:I, s:O}",
-          "code",
-          (json_int_t) TALER_EC_MERCHANT_TIP_PICKUP_DENOMINATION_UNKNOWN,
-          "exchange_code", (json_int_t) hr->ec,
-          "exchange_http_status", (json_int_t) hr->http_status,
-          "exchange_reply", hr->reply);
+        TALER_MHD_MAKE_JSON_PACK (
+          TALER_JSON_pack_ec (
+            TALER_EC_MERCHANT_TIP_PICKUP_DENOMINATION_UNKNOWN),
+          TMH_pack_exchange_reply (hr));
       MHD_resume_connection (pc->connection);
       TMH_trigger_daemon ();   /* we resumed, kick MHD */
       return;
@@ -772,6 +754,7 @@ TMH_post_tips_ID_pickup (const struct TMH_RequestHandler *rh,
                                           GNUNET_NO,
                                           &compute_total_requested,
                                           pc);
+    GNUNET_free (exchange_url);
     return MHD_YES;
   }
 
@@ -800,11 +783,10 @@ RETRY:
   }
   {
     struct GNUNET_CRYPTO_RsaSignature *sigs[GNUNET_NZL (pc->planchets_length)];
+
     memset (sigs,
             0,
-            sizeof (struct GNUNET_CRYPTO_RsaSignature *) * GNUNET_NZL (
-              pc->planchets_length));
-
+            sizeof (sigs));
     qs = TMH_db->lookup_pickup (TMH_db->cls,
                                 hc->instance->settings.id,
                                 &pc->tip_id,
@@ -817,29 +799,33 @@ RETRY:
       qs = GNUNET_DB_STATUS_SUCCESS_ONE_RESULT;
     if (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT == qs)
     {
+      bool rollback = false;
+
       for (unsigned int i = 0; i< pc->planchets_length; i++)
       {
-        if (NULL == sigs[i])
+        if (NULL != sigs[i])
+          continue;
+        if (! rollback)
         {
-          try_withdraw (pc,
-                        exchange_url,
-                        &pc->planchets[i],
-                        i);
-          qs = GNUNET_DB_STATUS_SUCCESS_NO_RESULTS;
+          TMH_db->rollback (TMH_db->cls);
+          MHD_suspend_connection (connection);
+          GNUNET_CONTAINER_DLL_insert (pc_head,
+                                       pc_tail,
+                                       pc);
+          pc->tt = GNUNET_SCHEDULER_add_delayed (EXCHANGE_TIMEOUT,
+                                                 &do_timeout,
+                                                 pc);
+          rollback = true;
         }
+        try_withdraw (pc,
+                      exchange_url,
+                      &pc->planchets[i],
+                      i);
       }
-      if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
-      {
-        MHD_suspend_connection (connection);
-        GNUNET_CONTAINER_DLL_insert (pc_head,
-                                     pc_tail,
-                                     pc);
-        pc->tt = GNUNET_SCHEDULER_add_delayed (EXCHANGE_TIMEOUT,
-                                               &do_timeout,
-                                               pc);
-        TMH_db->rollback (TMH_db->cls);
+      GNUNET_free (exchange_url);
+      if (rollback)
         return MHD_YES;
-      }
+      /* we got _all_ signatures, can continue! */
     }
     if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS != qs)
     {
@@ -860,22 +846,20 @@ RETRY:
             GNUNET_assert (0 ==
                            json_array_append_new (
                              blind_sigs,
-                             json_pack ("{s:o}",
-                                        "blind_sig",
-                                        GNUNET_JSON_from_rsa_signature (
-                                          sigs[i]))));
+                             GNUNET_JSON_PACK (
+                               GNUNET_JSON_pack_rsa_signature ("blind_sig",
+                                                               sigs[i]))));
             GNUNET_CRYPTO_rsa_signature_free (sigs[i]);
           }
-          return TALER_MHD_reply_json_pack (
+          return TALER_MHD_REPLY_JSON_PACK (
             connection,
             MHD_HTTP_OK,
-            "{s:o}",
-            "blind_sigs", blind_sigs);
+            GNUNET_JSON_pack_array_steal ("blind_sigs",
+                                          blind_sigs));
         }
         break;
       case GNUNET_DB_STATUS_SOFT_ERROR:
         goto RETRY;
-        break;
       case GNUNET_DB_STATUS_HARD_ERROR:
         ec = TALER_EC_GENERIC_DB_FETCH_FAILED;
         response_code = MHD_HTTP_INTERNAL_SERVER_ERROR;
@@ -902,12 +886,19 @@ RETRY:
                            &exchange_url,
                            &pc->reserve_priv);
   if (GNUNET_DB_STATUS_SOFT_ERROR == qs)
+  {
+    TMH_db->rollback (TMH_db->cls);
     goto RETRY;
+  }
   if (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT != qs)
+  {
+    TMH_db->rollback (TMH_db->cls);
     return reply_lookup_tip_failed (connection,
                                     qs);
-  if (0 == GNUNET_TIME_absolute_get_remaining (expiration).rel_value_us)
+  }
+  if (GNUNET_TIME_absolute_is_past (expiration))
   {
+    GNUNET_free (exchange_url);
     TMH_db->rollback (TMH_db->cls);
     return TALER_MHD_reply_with_error (connection,
                                        MHD_HTTP_GONE,
@@ -919,7 +910,8 @@ RETRY:
                              &total_authorized,
                              &total_picked_up))
   {
-    GNUNET_break (0);
+    GNUNET_free (exchange_url);
+    GNUNET_break_op (0);
     TMH_db->rollback (TMH_db->cls);
     return TALER_MHD_reply_with_error (connection,
                                        MHD_HTTP_INTERNAL_SERVER_ERROR,
@@ -931,7 +923,9 @@ RETRY:
       TALER_amount_cmp (&total_remaining,
                         &pc->total_requested))
   {
-    GNUNET_break (0);
+    /* total_remaining < pc->total_requested */
+    GNUNET_free (exchange_url);
+    GNUNET_break_op (0);
     TMH_db->rollback (TMH_db->cls);
     return TALER_MHD_reply_with_error (connection,
                                        MHD_HTTP_BAD_REQUEST,
@@ -954,6 +948,7 @@ RETRY:
     TMH_db->rollback (TMH_db->cls);
     if (GNUNET_DB_STATUS_SOFT_ERROR == qs)
       goto RETRY;
+    GNUNET_free (exchange_url);
     return TALER_MHD_reply_with_error (connection,
                                        MHD_HTTP_INTERNAL_SERVER_ERROR,
                                        TALER_EC_GENERIC_DB_STORE_FAILED,
@@ -965,6 +960,7 @@ RETRY:
     TMH_db->rollback (TMH_db->cls);
     if (GNUNET_DB_STATUS_SOFT_ERROR == qs)
       goto RETRY;
+    GNUNET_free (exchange_url);
     return TALER_MHD_reply_with_error (connection,
                                        MHD_HTTP_INTERNAL_SERVER_ERROR,
                                        TALER_EC_GENERIC_DB_COMMIT_FAILED,
@@ -981,5 +977,6 @@ RETRY:
                   &pc->planchets[i],
                   i);
   }
+  GNUNET_free (exchange_url);
   return MHD_YES;
 }
