@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2017, 2019, 2020 Taler Systems SA
+  (C) 2017-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
@@ -25,6 +25,7 @@
 #include <taler/taler_json_lib.h>
 #include "taler-merchant-httpd_mhd.h"
 #include "taler-merchant-httpd_exchanges.h"
+#include "taler-merchant-httpd_helper.h"
 #include "taler-merchant-httpd_private-get-orders.h"
 
 
@@ -337,28 +338,19 @@ gorc_report (struct GetOrderRequestContext *gorc,
     GNUNET_assert (0 ==
                    json_array_append_new (
                      gorc->wire_reports,
-                     json_pack ("{s:I, s:s, s:I, s:I, s:o }",
-                                "code",
-                                (json_int_t) ec,
-                                "hint",
-                                TALER_ErrorCode_get_hint (ec),
-                                "exchange_ec",
-                                (json_int_t) exchange_hr->ec,
-                                "exchange_hc",
-                                (json_int_t) exchange_hr->http_status,
-                                "coin_pub",
-                                GNUNET_JSON_from_data_auto (coin_pub))));
+                     GNUNET_JSON_PACK (
+                       TALER_JSON_pack_ec (ec),
+                       TMH_pack_exchange_reply (exchange_hr),
+                       GNUNET_JSON_pack_data_auto ("coin_pub",
+                                                   coin_pub))));
   else
     GNUNET_assert (0 ==
                    json_array_append_new (
                      gorc->wire_reports,
-                     json_pack ("{s:I, s:s, s:o }",
-                                "code",
-                                (json_int_t) ec,
-                                "hint",
-                                TALER_ErrorCode_get_hint (ec),
-                                "coin_pub",
-                                GNUNET_JSON_from_data_auto (coin_pub))));
+                     GNUNET_JSON_PACK (
+                       TALER_JSON_pack_ec (ec),
+                       GNUNET_JSON_pack_data_auto ("coin_pub",
+                                                   coin_pub))));
 }
 
 
@@ -657,13 +649,13 @@ process_refunds_cb (void *cls,
   GNUNET_assert (0 ==
                  json_array_append_new (
                    gorc->refund_details,
-                   json_pack ("{s:o, s:o, s:s}",
-                              "amount",
-                              TALER_JSON_from_amount (refund_amount),
-                              "timestamp",
-                              GNUNET_JSON_from_time_abs (timestamp),
-                              "reason",
-                              reason)));
+                   GNUNET_JSON_PACK (
+                     TALER_JSON_pack_amount ("amount",
+                                             refund_amount),
+                     GNUNET_JSON_pack_time_abs ("timestamp",
+                                                timestamp),
+                     GNUNET_JSON_pack_string ("reason",
+                                              reason))));
   /* For refunded coins, we are not charged deposit fees, so subtract those
      again */
   for (struct TransferQuery *tq = gorc->tq_head;
@@ -734,17 +726,17 @@ process_transfer_details (void *cls,
   GNUNET_assert (0 ==
                  json_array_append_new (
                    wire_details,
-                   json_pack ("{s:o, s:s, s:o, s:o, s:b}",
-                              "wtid",
-                              GNUNET_JSON_from_data_auto (wtid),
-                              "exchange_url",
-                              exchange_url,
-                              "amount",
-                              TALER_JSON_from_amount (&wired),
-                              "execution_time",
-                              GNUNET_JSON_from_time_abs (execution_time_round),
-                              "confirmed",
-                              transfer_confirmed)));
+                   GNUNET_JSON_PACK (
+                     GNUNET_JSON_pack_data_auto ("wtid",
+                                                 wtid),
+                     GNUNET_JSON_pack_string ("exchange_url",
+                                              exchange_url),
+                     TALER_JSON_pack_amount ("amount",
+                                             &wired),
+                     GNUNET_JSON_pack_time_abs ("execution_time",
+                                                execution_time_round),
+                     GNUNET_JSON_pack_bool ("confirmed",
+                                            transfer_confirmed))));
 }
 
 
@@ -759,6 +751,8 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
   bool wired;
   bool order_only = false;
   struct TALER_ClaimTokenP claim_token = { 0 };
+  const char *summary;
+  struct GNUNET_TIME_Absolute timestamp;
 
   if (NULL == gorc)
   {
@@ -845,7 +839,8 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
                                       hc->instance->settings.id,
                                       hc->infix,
                                       &gorc->contract_terms,
-                                      &gorc->order_serial);
+                                      &gorc->order_serial,
+                                      NULL);
   if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
   {
     order_only = true;
@@ -904,11 +899,20 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
       json_decref (ct);
     }
   }
-  /* extract the fulfillment URL and total amount from the contract terms! */
+  /* extract the fulfillment URL, total amount, summary and timestamp
+     from the contract terms! */
   {
     struct GNUNET_JSON_Specification spec[] = {
       TALER_JSON_spec_amount ("amount",
+                              TMH_currency,
                               &gorc->contract_amount),
+      GNUNET_JSON_spec_mark_optional (
+        GNUNET_JSON_spec_string ("fulfillment_url",
+                                 &gorc->fulfillment_url)),
+      GNUNET_JSON_spec_string ("summary",
+                               &summary),
+      TALER_JSON_spec_absolute_time ("timestamp",
+                                     &timestamp),
       GNUNET_JSON_spec_end ()
     };
 
@@ -924,21 +928,7 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
         TALER_EC_MERCHANT_GENERIC_DB_CONTRACT_CONTENT_INVALID,
         hc->infix);
     }
-    if (0 !=
-        strcasecmp (TMH_currency,
-                    gorc->contract_amount.currency))
-    {
-      GNUNET_break (0);
-      return TALER_MHD_reply_with_error (
-        connection,
-        MHD_HTTP_INTERNAL_SERVER_ERROR,
-        TALER_EC_MERCHANT_GENERIC_DB_CONTRACT_CONTENT_INVALID,
-        gorc->contract_amount.currency);
-    }
   }
-  gorc->fulfillment_url
-    = json_string_value (json_object_get (gorc->contract_terms,
-                                          "fulfillment_url"));
   if (! order_only)
   {
     if (GNUNET_OK !=
@@ -1036,19 +1026,36 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
                                                     hc->instance->settings.id,
                                                     &claim_token,
                                                     NULL);
-      ret = TALER_MHD_reply_json_pack (connection,
-                                       MHD_HTTP_OK,
-                                       "{s:s, s:s, s:s, s:s, s:s}",
-                                       "taler_pay_uri",
-                                       taler_pay_uri,
-                                       "order_status_url",
-                                       order_status_url,
-                                       "order_status",
-                                       "unpaid",
-                                       "already_paid_order_id",
-                                       already_paid_order_id,
-                                       "already_paid_fulfillment_url",
-                                       gorc->fulfillment_url);
+      if ( (NULL == taler_pay_uri) ||
+           (NULL == order_status_url) )
+      {
+        GNUNET_break_op (0);
+        GNUNET_free (taler_pay_uri);
+        GNUNET_free (order_status_url);
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_BAD_REQUEST,
+                                           TALER_EC_GENERIC_HTTP_HEADERS_MALFORMED,
+                                           "host");
+      }
+      ret = TALER_MHD_REPLY_JSON_PACK (
+        connection,
+        MHD_HTTP_OK,
+        GNUNET_JSON_pack_string ("taler_pay_uri",
+                                 taler_pay_uri),
+        GNUNET_JSON_pack_string ("order_status_url",
+                                 order_status_url),
+        GNUNET_JSON_pack_string ("order_status",
+                                 "unpaid"),
+        GNUNET_JSON_pack_string ("already_paid_order_id",
+                                 already_paid_order_id),
+        GNUNET_JSON_pack_string ("already_paid_fulfillment_url",
+                                 gorc->fulfillment_url),
+        TALER_JSON_pack_amount ("total_amount",
+                                &gorc->contract_amount),
+        GNUNET_JSON_pack_string ("summary",
+                                 summary),
+        GNUNET_JSON_pack_time_abs ("creation_time",
+                                   timestamp));
       GNUNET_free (taler_pay_uri);
       GNUNET_free (already_paid_order_id);
       return ret;
@@ -1061,16 +1068,29 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
   if ( (! paid) &&
        (! order_only) )
   {
+    if (GNUNET_TIME_absolute_is_future (gorc->sc.long_poll_timeout))
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                  "Suspending GET /private/orders/%s\n",
+                  hc->infix);
+      TMH_long_poll_suspend (hc->infix,
+                             gorc->session_id,
+                             gorc->fulfillment_url,
+                             hc->instance,
+                             &gorc->sc,
+                             NULL);
+      return MHD_YES;
+    }
     GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
                 "Order %s claimed but not paid yet\n",
                 hc->infix);
-    return TALER_MHD_reply_json_pack (connection,
-                                      MHD_HTTP_OK,
-                                      "{s:O, s:s}",
-                                      "contract_terms",
-                                      gorc->contract_terms,
-                                      "order_status",
-                                      "claimed");
+    return TALER_MHD_REPLY_JSON_PACK (
+      connection,
+      MHD_HTTP_OK,
+      GNUNET_JSON_pack_object_incref ("contract_terms",
+                                      gorc->contract_terms),
+      GNUNET_JSON_pack_string ("order_status",
+                               "claimed"));
   }
   if (paid &&
       (! wired) &&
@@ -1079,10 +1099,10 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
     /* suspend connection, wait for exchange to check wire transfer status there */
     gorc->transfer_status_requested = false;   /* only try ONCE */
     GNUNET_assert (GNUNET_OK ==
-                   TALER_amount_get_zero (TMH_currency,
+                   TALER_amount_set_zero (TMH_currency,
                                           &gorc->deposits_total));
     GNUNET_assert (GNUNET_OK ==
-                   TALER_amount_get_zero (TMH_currency,
+                   TALER_amount_set_zero (TMH_currency,
                                           &gorc->deposit_fees_total));
     TMH_db->lookup_deposits_by_order (TMH_db->cls,
                                       gorc->order_serial,
@@ -1102,7 +1122,7 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
   }
 
   if ( (! paid) &&
-       (0 != gorc->sc.long_poll_timeout.abs_value_us) )
+       (GNUNET_TIME_absolute_is_future (gorc->sc.long_poll_timeout)) )
   {
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                 "Suspending GET /private/orders/%s\n",
@@ -1134,15 +1154,21 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
                                                   hc->instance->settings.id,
                                                   &claim_token,
                                                   NULL);
-    ret = TALER_MHD_reply_json_pack (connection,
-                                     MHD_HTTP_OK,
-                                     "{s:s, s:s, s:s}",
-                                     "taler_pay_uri",
-                                     taler_pay_uri,
-                                     "order_status_url",
-                                     order_status_url,
-                                     "order_status",
-                                     "unpaid");
+    ret = TALER_MHD_REPLY_JSON_PACK (
+      connection,
+      MHD_HTTP_OK,
+      GNUNET_JSON_pack_string ("taler_pay_uri",
+                               taler_pay_uri),
+      GNUNET_JSON_pack_string ("order_status_url",
+                               order_status_url),
+      GNUNET_JSON_pack_string ("order_status",
+                               "unpaid"),
+      TALER_JSON_pack_amount ("total_amount",
+                              &gorc->contract_amount),
+      GNUNET_JSON_pack_string ("summary",
+                               summary),
+      GNUNET_JSON_pack_time_abs ("creation_time",
+                                 timestamp));
     GNUNET_free (taler_pay_uri);
     GNUNET_free (order_status_url);
     return ret;
@@ -1154,7 +1180,7 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
   /* Accumulate refunds, if any. */
   {
     GNUNET_assert (GNUNET_OK ==
-                   TALER_amount_get_zero (TMH_currency,
+                   TALER_amount_set_zero (TMH_currency,
                                           &gorc->refund_amount));
     qs = TMH_db->lookup_refunds_detailed (TMH_db->cls,
                                           hc->instance->settings.id,
@@ -1177,10 +1203,10 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
     char *order_status_url;
 
     GNUNET_assert (GNUNET_OK ==
-                   TALER_amount_get_zero (TMH_currency,
+                   TALER_amount_set_zero (TMH_currency,
                                           &gorc->deposits_total));
     GNUNET_assert (GNUNET_OK ==
-                   TALER_amount_get_zero (TMH_currency,
+                   TALER_amount_set_zero (TMH_currency,
                                           &gorc->deposit_fees_total));
     qs = TMH_db->lookup_transfer_details_by_order (TMH_db->cls,
                                                    gorc->order_serial,
@@ -1229,30 +1255,11 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
           TALER_amount_cmp (&expect_total,
                             &gorc->deposits_total))
       {
-        struct GNUNET_TIME_Absolute timestamp;
-
         /* expect_total <= gorc->deposits_total: good: we got paid */
         wired = true;
         qs = TMH_db->mark_order_wired (TMH_db->cls,
                                        gorc->order_serial);
         GNUNET_break (qs >= 0); /* just warn if transaction failed */
-        {
-          struct GNUNET_JSON_Specification spec[] = {
-            TALER_JSON_spec_absolute_time ("timestamp",
-                                           &timestamp),
-            GNUNET_JSON_spec_end ()
-          };
-          enum GNUNET_GenericReturnValue res;
-
-          res = TALER_MHD_parse_internal_json_data (connection,
-                                                    gorc->contract_terms,
-                                                    spec);
-          if (GNUNET_YES != res)
-          {
-            GNUNET_break (0);
-            return res;
-          }
-        }
         TMH_notify_order_change (hc->instance,
                                  hc->infix,
                                  true, /* paid */
@@ -1283,38 +1290,41 @@ TMH_private_get_orders_ID (const struct TMH_RequestHandler *rh,
                                    h_contract);
     }
 
-    ret = TALER_MHD_reply_json_pack (
+    ret = TALER_MHD_REPLY_JSON_PACK (
       connection,
       MHD_HTTP_OK,
-      "{s:o, s:I, s:I, s:o, s:O,"
-      " s:s, s:b, s:b, s:b, s:o,"
-      " s:o, s:o, s:s}",
-      "wire_reports",
-      gorc->wire_reports,
-      "exchange_ec",
-      (json_int_t) gorc->exchange_ec,
-      "exchange_hc",
-      (json_int_t) gorc->exchange_hc,
-      "deposit_total",
-      TALER_JSON_from_amount (&gorc->deposits_total),
-      "contract_terms",
-      gorc->contract_terms,
-      "order_status",
-      "paid",
-      "refunded",
-      gorc->refunded,
-      "wired",
-      wired,
-      "refund_pending",
-      gorc->refund_pending,
-      "refund_amount",
-      TALER_JSON_from_amount (&gorc->refund_amount),
-      "wire_details",
-      gorc->wire_details,
-      "refund_details",
-      gorc->refund_details,
-      "order_status_url",
-      order_status_url);
+      GNUNET_JSON_pack_array_steal ("wire_reports",
+                                    gorc->wire_reports),
+      GNUNET_JSON_pack_uint64 ("exchange_code",
+                               gorc->exchange_ec),
+      GNUNET_JSON_pack_uint64 ("exchange_http_status",
+                               gorc->exchange_hc),
+      /* legacy: */
+      GNUNET_JSON_pack_uint64 ("exchange_ec",
+                               gorc->exchange_ec),
+      /* legacy: */
+      GNUNET_JSON_pack_uint64 ("exchange_hc",
+                               gorc->exchange_hc),
+      TALER_JSON_pack_amount ("deposit_total",
+                              &gorc->deposits_total),
+      GNUNET_JSON_pack_object_incref ("contract_terms",
+                                      gorc->contract_terms),
+      GNUNET_JSON_pack_string ("order_status",
+                               "paid"),
+      GNUNET_JSON_pack_bool ("refunded",
+                             gorc->refunded),
+      GNUNET_JSON_pack_bool ("wired",
+                             wired),
+      GNUNET_JSON_pack_bool ("refund_pending",
+                             gorc->refund_pending),
+      TALER_JSON_pack_amount ("refund_amount",
+                              &gorc->refund_amount),
+      GNUNET_JSON_pack_array_steal ("wire_details",
+                                    gorc->wire_details),
+      GNUNET_JSON_pack_array_steal ("refund_details",
+                                    gorc->refund_details),
+      GNUNET_JSON_pack_string ("order_status_url",
+                               order_status_url));
     GNUNET_free (order_status_url);
     gorc->wire_details = NULL;
     gorc->wire_reports = NULL;
