@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2019, 2020 Taler Systems SA
+  (C) 2019-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -24,30 +24,13 @@
 
 
 /**
- * Stores state for adding an order to the array for the response.
+ * Sensible bound on TALER_MERCHANTDB_OrderFilter.delta
  */
-struct AddOrderState
-{
-  /**
-   * The array of orders.
-   */
-  json_t *pa;
-
-  /**
-   * The name of the instance we are querying for.
-   */
-  const char *instance_id;
-
-  /**
-   * The result after adding the orders (#TALER_EC_NONE for okay, anything else for an error).
-   */
-  enum TALER_ErrorCode result;
-
-};
+#define MAX_DELTA 1024
 
 
 /**
- * A pending GET /orders request that is in long polling mode.
+ * A pending GET /orders request.
  */
 struct TMH_PendingOrder
 {
@@ -84,16 +67,30 @@ struct TMH_PendingOrder
    */
   struct GNUNET_TIME_Absolute long_poll_timeout;
 
-  /**
-   * State for adding orders. The array `pa` must be
-   * json_decref()'ed when done with the `struct TMH_PendingOrder`!
-   */
-  struct AddOrderState *aos;
-
   /**
    * Filter to apply.
    */
   struct TALER_MERCHANTDB_OrderFilter of;
+
+  /**
+   * The array of orders.
+   */
+  json_t *pa;
+
+  /**
+   * The name of the instance we are querying for.
+   */
+  const char *instance_id;
+
+  /**
+   * The result after adding the orders (#TALER_EC_NONE for okay, anything else for an error).
+   */
+  enum TALER_ErrorCode result;
+
+  /**
+   * Is the structure in the DLL
+   */
+  bool in_dll;
 };
 
 
@@ -121,14 +118,14 @@ TMH_force_get_orders_resume (struct TMH_MerchantInstance *mi)
 
   while (NULL != (po = mi->po_head))
   {
+    GNUNET_assert (po->in_dll);
     GNUNET_CONTAINER_DLL_remove (mi->po_head,
                                  mi->po_tail,
                                  po);
     GNUNET_assert (po ==
                    GNUNET_CONTAINER_heap_remove_root (order_timeout_heap));
     MHD_resume_connection (po->con);
-    json_decref (po->aos->pa);
-    GNUNET_free (po);
+    po->in_dll = false;
   }
   if (NULL != order_timeout_task)
   {
@@ -166,9 +163,7 @@ order_timeout (void *cls)
       order_timeout_heap = NULL;
       return;
     }
-    if  (0 !=
-         GNUNET_TIME_absolute_get_remaining (
-           po->long_poll_timeout).rel_value_us)
+    if (GNUNET_TIME_absolute_is_future (po->long_poll_timeout))
       break;
     GNUNET_assert (po ==
                    GNUNET_CONTAINER_heap_remove_root (order_timeout_heap));
@@ -176,13 +171,13 @@ order_timeout (void *cls)
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                 "Resuming long polled job due to timeout\n");
     mi = po->mi;
+    GNUNET_assert (po->in_dll);
     GNUNET_CONTAINER_DLL_remove (mi->po_head,
                                  mi->po_tail,
                                  po);
-    json_decref (po->aos->pa);
+    po->in_dll = false;
     MHD_resume_connection (po->con);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
-    GNUNET_free (po);
   }
   order_timeout_task = GNUNET_SCHEDULER_add_at (po->long_poll_timeout,
                                                 &order_timeout,
@@ -199,9 +194,21 @@ order_timeout (void *cls)
 static void
 cleanup (void *ctx)
 {
-  struct AddOrderState *aos = ctx;
-  json_decref (aos->pa);
-  GNUNET_free (aos);
+  struct TMH_PendingOrder *po = ctx;
+
+  if (po->in_dll)
+  {
+    struct TMH_MerchantInstance *mi = po->mi;
+
+    GNUNET_CONTAINER_DLL_remove (mi->po_head,
+                                 mi->po_tail,
+                                 po);
+  }
+  if (NULL != po->hn)
+    GNUNET_assert (po ==
+                   GNUNET_CONTAINER_heap_remove_node (po->hn));
+  json_decref (po->pa);
+  GNUNET_free (po);
 }
 
 
@@ -253,28 +260,28 @@ add_order (void *cls,
            uint64_t order_serial,
            struct GNUNET_TIME_Absolute creation_time)
 {
-  struct AddOrderState *aos = cls;
+  struct TMH_PendingOrder *po = cls;
   json_t *contract_terms;
   struct GNUNET_HashCode h_contract_terms;
   enum GNUNET_DB_QueryStatus qs;
+  const char *summary;
   bool refundable = false;
   bool paid;
+  struct TALER_Amount order_amount;
 
+  qs = TMH_db->lookup_order_status (TMH_db->cls,
+                                    po->instance_id,
+                                    order_id,
+                                    &h_contract_terms,
+                                    &paid);
+  /* qs == 0: contract terms don't exist, so the order cannot be paid. */
+  if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
+    paid = false;
+  if (qs < 0)
   {
-    qs = TMH_db->lookup_order_status (TMH_db->cls,
-                                      aos->instance_id,
-                                      order_id,
-                                      &h_contract_terms,
-                                      &paid);
-    /* qs == 0: contract terms don't exist, so the order cannot be paid. */
-    if (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs)
-      paid = false;
-    if (qs < 0)
-    {
-      GNUNET_break (0);
-      aos->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
-      return;
-    }
+    GNUNET_break (0);
+    po->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
+    return;
   }
 
   if (paid)
@@ -284,17 +291,18 @@ add_order (void *cls,
     uint64_t os;
 
     qs = TMH_db->lookup_contract_terms (TMH_db->cls,
-                                        aos->instance_id,
+                                        po->instance_id,
                                         order_id,
                                         &contract_terms,
-                                        &os);
+                                        &os,
+                                        NULL);
   }
   else
   {
     struct GNUNET_HashCode unused;
 
     qs = TMH_db->lookup_order (TMH_db->cls,
-                               aos->instance_id,
+                               po->instance_id,
                                order_id,
                                NULL,
                                &unused,
@@ -304,22 +312,21 @@ add_order (void *cls,
   if (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT != qs)
   {
     GNUNET_break (0);
-    aos->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
+    po->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
     json_decref (contract_terms);
     return;
   }
 
   {
-    struct TALER_Amount order_amount;
     struct GNUNET_TIME_Absolute rd;
-
-    struct GNUNET_TIME_Absolute now = GNUNET_TIME_absolute_get ();
-
     struct GNUNET_JSON_Specification spec[] = {
       TALER_JSON_spec_amount ("amount",
+                              TMH_currency,
                               &order_amount),
       GNUNET_JSON_spec_absolute_time ("refund_deadline",
                                       &rd),
+      GNUNET_JSON_spec_string ("summary",
+                               &summary),
       GNUNET_JSON_spec_end ()
     };
 
@@ -329,28 +336,28 @@ add_order (void *cls,
                            NULL, NULL))
     {
       GNUNET_break (0);
-      aos->result = TALER_EC_MERCHANT_GENERIC_DB_CONTRACT_CONTENT_INVALID;
+      po->result = TALER_EC_MERCHANT_GENERIC_DB_CONTRACT_CONTENT_INVALID;
       json_decref (contract_terms);
       return;
     }
 
-    if ((now.abs_value_us <= rd.abs_value_us) &&
+    if (GNUNET_TIME_absolute_is_future (rd) &&
         paid)
     {
       struct TALER_Amount refund_amount;
 
       GNUNET_assert (GNUNET_OK ==
-                     TALER_amount_get_zero (TMH_currency,
+                     TALER_amount_set_zero (TMH_currency,
                                             &refund_amount));
       qs = TMH_db->lookup_refunds_detailed (TMH_db->cls,
-                                            aos->instance_id,
+                                            po->instance_id,
                                             &h_contract_terms,
                                             &process_refunds_cb,
                                             &refund_amount);
       if (0 > qs)
       {
         GNUNET_break (0);
-        aos->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
+        po->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
         json_decref (contract_terms);
         return;
       }
@@ -362,25 +369,22 @@ add_order (void *cls,
 
   GNUNET_assert (0 ==
                  json_array_append_new (
-                   aos->pa,
-                   json_pack (
-                     "{s:s, s:I, s:o, s:O, s:O, s:b, s:b}",
-                     "order_id",
-                     order_id,
-                     "row_id",
-                     (json_int_t) order_serial,
-                     "timestamp",
-                     GNUNET_JSON_from_time_abs (creation_time),
-                     "amount",
-                     json_object_get (contract_terms,
-                                      "amount"),
-                     "summary",
-                     json_object_get (contract_terms,
-                                      "summary"),
-                     "refundable",
-                     refundable,
-                     "paid",
-                     paid)));
+                   po->pa,
+                   GNUNET_JSON_PACK (
+                     GNUNET_JSON_pack_string ("order_id",
+                                              order_id),
+                     GNUNET_JSON_pack_uint64 ("row_id",
+                                              order_serial),
+                     GNUNET_JSON_pack_time_abs ("timestamp",
+                                                creation_time),
+                     TALER_JSON_pack_amount ("amount",
+                                             &order_amount),
+                     GNUNET_JSON_pack_string ("summary",
+                                              summary),
+                     GNUNET_JSON_pack_bool ("refundable",
+                                            refundable),
+                     GNUNET_JSON_pack_bool ("paid",
+                                            paid))));
   json_decref (contract_terms);
 }
 
@@ -436,19 +440,19 @@ TMH_notify_order_change (struct TMH_MerchantInstance *mi,
         continue;
       po->of.delta++;
     }
-    add_order (po->aos,
+    add_order (po,
                order_id,
                order_serial_id,
                date);
+    GNUNET_assert (po->in_dll);
     GNUNET_CONTAINER_DLL_remove (mi->po_head,
                                  mi->po_tail,
                                  po);
+    po->in_dll = false;
     GNUNET_assert (po ==
                    GNUNET_CONTAINER_heap_remove_node (po->hn));
     MHD_resume_connection (po->con);
     TMH_trigger_daemon ();   /* we resumed, kick MHD */
-    json_decref (po->aos->pa);
-    GNUNET_free (po);
   }
 }
 
@@ -466,28 +470,27 @@ TMH_private_get_orders (const struct TMH_RequestHandler *rh,
                         struct MHD_Connection *connection,
                         struct TMH_HandlerContext *hc)
 {
-  struct AddOrderState *aos;
+  struct TMH_PendingOrder *po = hc->ctx;
   enum GNUNET_DB_QueryStatus qs;
   struct TALER_MERCHANTDB_OrderFilter of;
 
-  if (NULL != hc->ctx)
+  if (NULL != po)
   {
     /* resumed from long-polling, return answer we already have
        in 'hc->ctx' */
-    struct AddOrderState *aos = hc->ctx;
-
-    if (TALER_EC_NONE != aos->result)
+    if (TALER_EC_NONE != po->result)
     {
       GNUNET_break (0);
       return TALER_MHD_reply_with_error (connection,
                                          MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                         aos->result,
+                                         po->result,
                                          NULL);
     }
-    return TALER_MHD_reply_json_pack (connection,
-                                      MHD_HTTP_OK,
-                                      "{s:O}",
-                                      "orders", aos->pa);
+    return TALER_MHD_REPLY_JSON_PACK (
+      connection,
+      MHD_HTTP_OK,
+      GNUNET_JSON_pack_array_incref ("orders",
+                                     po->pa));
   }
 
   if (! (TALER_arg_to_yna (connection,
@@ -514,33 +517,6 @@ TMH_private_get_orders (const struct TMH_RequestHandler *rh,
                                        MHD_HTTP_BAD_REQUEST,
                                        TALER_EC_GENERIC_PARAMETER_MALFORMED,
                                        "wired");
-  {
-    const char *start_row_str;
-
-    start_row_str = MHD_lookup_connection_value (connection,
-                                                 MHD_GET_ARGUMENT_KIND,
-                                                 "start");
-    if (NULL == start_row_str)
-    {
-      of.start_row = INT64_MAX;
-    }
-    else
-    {
-      char dummy[2];
-      unsigned long long ull;
-
-      if (1 !=
-          sscanf (start_row_str,
-                  "%llu%1s",
-                  &ull,
-                  dummy))
-        return TALER_MHD_reply_with_error (connection,
-                                           MHD_HTTP_BAD_REQUEST,
-                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
-                                           "date");
-      of.start_row = (uint64_t) ull;
-    }
-  }
   {
     const char *delta_str;
 
@@ -553,28 +529,40 @@ TMH_private_get_orders (const struct TMH_RequestHandler *rh,
     }
     else
     {
-      char dummy[2];
+      char dummy;
       long long ll;
 
       if (1 !=
           sscanf (delta_str,
-                  "%lld%1s",
+                  "%lld%c",
                   &ll,
-                  dummy))
+                  &dummy))
+      {
+        GNUNET_break_op (0);
         return TALER_MHD_reply_with_error (connection,
                                            MHD_HTTP_BAD_REQUEST,
                                            TALER_EC_GENERIC_PARAMETER_MALFORMED,
                                            "delta");
-      of.delta = (uint64_t) ll;
+      }
+      of.delta = (int64_t) ll;
+      if ( (-MAX_DELTA > of.delta) ||
+           (of.delta > MAX_DELTA) )
+      {
+        GNUNET_break_op (0);
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_BAD_REQUEST,
+                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
+                                           "delta");
+      }
     }
   }
   {
-    const char *date_str;
+    const char *date_ms_str;
 
-    date_str = MHD_lookup_connection_value (connection,
-                                            MHD_GET_ARGUMENT_KIND,
-                                            "date");
-    if (NULL == date_str)
+    date_ms_str = MHD_lookup_connection_value (connection,
+                                               MHD_GET_ARGUMENT_KIND,
+                                               "date_ms");
+    if (NULL == date_ms_str)
     {
       if (of.delta > 0)
         of.date = GNUNET_TIME_UNIT_ZERO_ABS;
@@ -583,13 +571,69 @@ TMH_private_get_orders (const struct TMH_RequestHandler *rh,
     }
     else
     {
-      if (GNUNET_OK !=
-          GNUNET_STRINGS_fancy_time_to_absolute (date_str,
-                                                 &of.date))
+      char dummy;
+      unsigned long long ll;
+
+      if (1 !=
+          sscanf (date_ms_str,
+                  "%llu%c",
+                  &ll,
+                  &dummy))
+      {
+        GNUNET_break_op (0);
         return TALER_MHD_reply_with_error (connection,
                                            MHD_HTTP_BAD_REQUEST,
                                            TALER_EC_GENERIC_PARAMETER_MALFORMED,
-                                           "date");
+                                           "date_ms");
+      }
+
+      of.date = GNUNET_TIME_absolute_from_ms (ll);
+      if (GNUNET_TIME_absolute_is_never (of.date))
+      {
+        GNUNET_break_op (0);
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_BAD_REQUEST,
+                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
+                                           "date_ms");
+      }
+    }
+  }
+  {
+    const char *start_row_str;
+
+    start_row_str = MHD_lookup_connection_value (connection,
+                                                 MHD_GET_ARGUMENT_KIND,
+                                                 "start");
+    if (NULL == start_row_str)
+    {
+      if (of.delta > 0)
+        of.start_row = 0;
+      else
+        of.start_row = INT64_MAX;
+    }
+    else
+    {
+      char dummy;
+      unsigned long long ull;
+
+      if (1 !=
+          sscanf (start_row_str,
+                  "%llu%c",
+                  &ull,
+                  &dummy))
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_BAD_REQUEST,
+                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
+                                           "start");
+      of.start_row = (uint64_t) ull;
+      if (INT64_MAX < of.start_row)
+      {
+        GNUNET_break_op (0);
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_BAD_REQUEST,
+                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
+                                           "start");
+      }
     }
   }
   {
@@ -604,69 +648,73 @@ TMH_private_get_orders (const struct TMH_RequestHandler *rh,
     }
     else
     {
-      char dummy[2];
+      char dummy;
       unsigned long long ull;
 
       if (1 !=
           sscanf (timeout_ms_str,
-                  "%lld%1s",
+                  "%lld%c",
                   &ull,
-                  dummy))
+                  &dummy))
         return TALER_MHD_reply_with_error (connection,
                                            MHD_HTTP_BAD_REQUEST,
                                            TALER_EC_GENERIC_PARAMETER_MALFORMED,
                                            "timeout_ms");
       of.timeout = GNUNET_TIME_relative_multiply (GNUNET_TIME_UNIT_MILLISECONDS,
                                                   ull);
+      if (GNUNET_TIME_relative_is_forever (of.timeout))
+      {
+        GNUNET_break_op (0);
+        return TALER_MHD_reply_with_error (connection,
+                                           MHD_HTTP_BAD_REQUEST,
+                                           TALER_EC_GENERIC_PARAMETER_MALFORMED,
+                                           "timeout_ms");
+      }
+    }
+
+    if ( (0 >= of.delta) &&
+         (! GNUNET_TIME_relative_is_zero (of.timeout)) )
+    {
+      GNUNET_break_op (0);
+      of.timeout = GNUNET_TIME_UNIT_ZERO;
     }
   }
 
-  aos = GNUNET_new (struct AddOrderState);
-  GNUNET_assert (NULL != aos);
-  aos->pa = json_array ();
-  aos->instance_id = hc->instance->settings.id;
-  aos->result = TALER_EC_NONE;
-  GNUNET_assert (NULL != aos->pa);
+  po = GNUNET_new (struct TMH_PendingOrder);
+  hc->ctx = po;
+  hc->cc = &cleanup;
+  po->con = connection;
+  po->pa = json_array ();
+  GNUNET_assert (NULL != po->pa);
+  po->instance_id = hc->instance->settings.id;
+  po->mi = hc->instance;
+
+  qs = TMH_db->lookup_orders (TMH_db->cls,
+                              po->instance_id,
+                              &of,
+                              &add_order,
+                              po);
+  if (0 > qs)
   {
-    qs = TMH_db->lookup_orders (TMH_db->cls,
-                                hc->instance->settings.id,
-                                &of,
-                                &add_order,
-                                aos);
-    if (0 > qs)
-    {
-      aos->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
-    }
-    if (TALER_EC_NONE != aos->result)
-    {
-      enum TALER_ErrorCode aos_result = aos->result;
-
-      GNUNET_break (0);
-      json_decref (aos->pa);
-      GNUNET_free (aos);
-      return TALER_MHD_reply_with_error (connection,
-                                         MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                         aos_result,
-                                         NULL);
-    }
+    po->result = TALER_EC_GENERIC_DB_FETCH_FAILED;
   }
-  if ( (0 == qs) &&
-       (of.timeout.rel_value_us > 0) )
+  if (TALER_EC_NONE != po->result)
+  {
+    GNUNET_break (0);
+    return TALER_MHD_reply_with_error (connection,
+                                       MHD_HTTP_INTERNAL_SERVER_ERROR,
+                                       po->result,
+                                       NULL);
+  }
+  if ( (GNUNET_DB_STATUS_SUCCESS_NO_RESULTS == qs) &&
+       (! GNUNET_TIME_relative_is_zero (of.timeout)) )
   {
     struct TMH_MerchantInstance *mi = hc->instance;
-    struct TMH_PendingOrder *po;
 
     /* setup timeout heap (if not yet exists) */
     if (NULL == order_timeout_heap)
       order_timeout_heap
         = GNUNET_CONTAINER_heap_create (GNUNET_CONTAINER_HEAP_ORDER_MIN);
-    hc->ctx = aos;
-    hc->cc = &cleanup;
-    po = GNUNET_new (struct TMH_PendingOrder);
-    po->mi = mi;
-    po->con = connection;
-    po->aos = aos;
-    json_incref (po->aos->pa);
     po->hn = GNUNET_CONTAINER_heap_insert (order_timeout_heap,
                                            po,
                                            po->long_poll_timeout.abs_value_us);
@@ -675,24 +723,26 @@ TMH_private_get_orders (const struct TMH_RequestHandler *rh,
     GNUNET_CONTAINER_DLL_insert (mi->po_head,
                                  mi->po_tail,
                                  po);
+    po->in_dll = true;
     MHD_suspend_connection (connection);
-    /* start timeout task */
-    po = GNUNET_CONTAINER_heap_peek (order_timeout_heap);
-    if (NULL != order_timeout_task)
-      GNUNET_SCHEDULER_cancel (order_timeout_task);
-    order_timeout_task = GNUNET_SCHEDULER_add_at (po->long_poll_timeout,
-                                                  &order_timeout,
-                                                  NULL);
+    {
+      struct TMH_PendingOrder *pot;
+
+      /* start timeout task */
+      pot = GNUNET_CONTAINER_heap_peek (order_timeout_heap);
+      if (NULL != order_timeout_task)
+        GNUNET_SCHEDULER_cancel (order_timeout_task);
+      order_timeout_task = GNUNET_SCHEDULER_add_at (pot->long_poll_timeout,
+                                                    &order_timeout,
+                                                    NULL);
+    }
     return MHD_YES;
   }
-  {
-    json_t *pa = aos->pa;
-    GNUNET_free (aos);
-    return TALER_MHD_reply_json_pack (connection,
-                                      MHD_HTTP_OK,
-                                      "{s:o}",
-                                      "orders", pa);
-  }
+  return TALER_MHD_REPLY_JSON_PACK (
+    connection,
+    MHD_HTTP_OK,
+    GNUNET_JSON_pack_array_incref ("orders",
+                                   po->pa));
 }
 
 
