@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015, 2016 GNUnet e.V. and INRIA
+  Copyright (C) 2014, 2015, 2016, 2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -15,7 +15,7 @@
 */
 
 /**
- * @file include/platform.h
+ * @file merchant/src/include/platform.h
  * @brief This file contains the includes and definitions which are used by the
  *        rest of the modules
  * @author Sree Harsha Totakura <sreeharsha@totakura.in>
@@ -55,6 +55,16 @@
    to "ancient" MHD releases. */
 #define MHD_NO_DEPRECATION 1
 
+
+/**
+ * Required prefix for the authorization header as per RFC 8959.
+ * (Follows RFC 6750 albeit technically violates RFC 7235, but
+ *  Mark Nottingham thinks this should be fixed by revising HTTP
+ *  spec (https://github.com/httpwg/http-core/issues/733))
+ */
+#define RFC_8959_PREFIX "secret-token:"
+
+
 #endif  /* PLATFORM_H_ */
 
 /* end of platform.h */
