@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015, 2016 INRIA
+  Copyright (C) 2014, 2015, 2016, 2020 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Lesser General Public License as published by the Free Software
@@ -15,7 +15,7 @@
 */
 
 /**
- * @file include/taler_merchantdb_lib.h
+ * @file taler_merchantdb_lib.h
  * @brief database helper functions used by the merchant backend
  * @author Sree Harsha Totakura <sreeharsha@totakura.in>
  */
@@ -33,11 +33,11 @@ struct TALER_MERCHANTDB_Plugin;
 /**
  * Connect to postgresql database
  *
- * @param[in,out] cfg the configuration handle
+ * @param cfg the configuration handle
  * @return connection to the database; NULL upon error
  */
 struct TALER_MERCHANTDB_Plugin *
-TALER_MERCHANTDB_plugin_load (struct GNUNET_CONFIGURATION_Handle *cfg);
+TALER_MERCHANTDB_plugin_load (const struct GNUNET_CONFIGURATION_Handle *cfg);
 
 
 /**
@@ -49,6 +49,16 @@ void
 TALER_MERCHANTDB_plugin_unload (struct TALER_MERCHANTDB_Plugin *dbh);
 
 
+/**
+ * Free members of @a pd, but not @a pd itself.
+ *
+ * @param[in] pd product details to clean up
+ */
+void
+TALER_MERCHANTDB_product_details_free (
+  struct TALER_MERCHANTDB_ProductDetails *pd);
+
+
 #endif  /* MERCHANT_DB_H */
 
 /* end of taler_merchantdb_lib.h */
