@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014-2020 Taler Systems SA
+  Copyright (C) 2014-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Lesser General Public License as published by the Free Software
@@ -60,6 +60,22 @@ struct TALER_MERCHANTDB_AccountDetails
 
 };
 
+/**
+ * Authentication settings for an instance.
+ */
+struct TALER_MERCHANTDB_InstanceAuthSettings
+{
+  /**
+   * Hash used for authentication.  All zero if authentication is off.
+   */
+  struct GNUNET_HashCode auth_hash;
+
+  /**
+   * Salt used to hash the "Authentication" header, the result must then
+   * match the @e auth_hash.
+   */
+  struct GNUNET_ShortHashCode auth_salt;
+};
 
 /**
  * General settings for an instance.
@@ -128,6 +144,7 @@ struct TALER_MERCHANTDB_InstanceSettings
  * @param merchant_pub public key of the instance
  * @param merchant_priv private key of the instance, NULL if not available
  * @param is general instance settings
+ * @param ias instance authentication settings
  * @param accounts_length length of the @a accounts array
  * @param accounts list of accounts of the merchant
  */
@@ -137,6 +154,7 @@ typedef void
   const struct TALER_MerchantPublicKeyP *merchant_pub,
   const struct TALER_MerchantPrivateKeyP *merchant_priv,
   const struct TALER_MERCHANTDB_InstanceSettings *is,
+  const struct TALER_MERCHANTDB_InstanceAuthSettings *ias,
   unsigned int accounts_length,
   const struct TALER_MERCHANTDB_AccountDetails accounts[]);
 
@@ -178,6 +196,11 @@ struct TALER_MERCHANTDB_ProductDetails
    */
   struct TALER_Amount price;
 
+  /**
+   * Base64-encoded product image, or an empty string.
+   */
+  char *image;
+
   /**
    * List of taxes the merchant pays for this product. Never NULL,
    * but can be an empty array.
@@ -201,11 +224,6 @@ struct TALER_MERCHANTDB_ProductDetails
    */
   uint64_t total_lost;
 
-  /**
-   * Base64-encoded product image, or an empty string.
-   */
-  json_t *image;
-
   /**
    * Identifies where the product is in stock, possibly an empty map.
    */
@@ -560,6 +578,8 @@ typedef void
  * @param committed_amount total of tips that the merchant committed to, but that were not
  *           picked up yet
  * @param active true if the reserve is still active (we have the private key)
+ * @param exchange_url base URL of the exchange hosting the reserve, NULL if not @a active
+ * @param payto_uri URI to use to fund the reserve, NULL if not @a active
  * @param tips_length length of the @a tips array
  * @param tips information about the tips created by this reserve
  */
@@ -573,6 +593,8 @@ typedef void
   const struct TALER_Amount *picked_up_amount,
   const struct TALER_Amount *committed_amount,
   bool active,
+  const char *exchange_url,
+  const char *payto_uri,
   unsigned int tips_length,
   const struct TALER_MERCHANTDB_TipDetails *tips);
 
@@ -669,19 +691,15 @@ struct TALER_MERCHANTDB_Plugin
   char *library_name;
 
   /**
-   * Connect to the database, possibly only to drop the tables. Called first
-   * to complete the initialization.
+   * Connect to the database.
    *
    * @param cls closure
-   * @param for_drop are we going to call drop_tables() next (and then
-   *         possibly again connect()?
    */
   int
-  (*connect) (void *cls,
-              bool for_drop);
+  (*connect) (void *cls);
 
   /**
-   * Drop merchant tables. Used for testcases.
+   * Drop merchant tables. Used for testcases and to reset the DB.
    *
    * @param cls closure
    * @return #GNUNET_OK upon success; #GNUNET_SYSERR upon failure
@@ -690,9 +708,58 @@ struct TALER_MERCHANTDB_Plugin
   (*drop_tables) (void *cls);
 
   /**
-   * Do a pre-flight check that we are not in an uncommitted transaction.
-   * If we are, try to commit the previous transaction and output a warning.
-   * Does not return anything, as we will continue regardless of the outcome.
+   * Initialize merchant tables
+   *
+   * @param cls closure
+   * @return #GNUNET_OK upon success; #GNUNET_SYSERR upon failure
+   */
+  int
+  (*create_tables) (void *cls);
+
+  /**
+   * Register callback to be invoked on events of type @a es.
+   *
+   * @param cls database context to use
+   * @param es specification of the event to listen for
+   * @param cb function to call when the event happens, possibly
+   *         multiple times (until cancel is invoked)
+   * @param cb_cls closure for @a cb
+   * @return handle useful to cancel the listener
+   */
+  struct GNUNET_DB_EventHandler *
+  (*event_listen)(void *cls,
+                  const struct GNUNET_DB_EventHeaderP *es,
+                  GNUNET_DB_EventCallback cb,
+                  void *cb_cls);
+
+  /**
+   * Stop notifications.
+   *
+   * @param eh handle to unregister.
+   */
+  void
+  (*event_listen_cancel)(struct GNUNET_DB_EventHandler *eh);
+
+
+  /**
+   * Notify all that listen on @a es of an event.
+   *
+   * @param cls database context to use
+   * @param es specification of the event to generate
+   * @param extra additional event data provided
+   * @param extra_size number of bytes in @a extra
+   */
+  void
+  (*event_notify)(void *cls,
+                  const struct GNUNET_DB_EventHeaderP *es,
+                  const void *extra,
+                  size_t extra_size);
+
+
+  /**
+   * Do a pre-flight check that we are not in an uncommitted transaction.  If
+   * we are, die.  Does not return anything, as we will continue regardless of
+   * the outcome.
    *
    * @param cls the `struct PostgresClosure` with the plugin-specific state
    */
@@ -755,6 +822,35 @@ struct TALER_MERCHANTDB_Plugin
                       TALER_MERCHANTDB_InstanceCallback cb,
                       void *cb_cls);
 
+  /**
+   * Lookup one of the instances this backend has configured.
+   *
+   * @param cls closure
+   * @param id ID of instance to look up
+   * @param active_only only find 'active' instances
+   * @param cb function to call on all instances found
+   * @param cb_cls closure for @a cb
+   */
+  enum GNUNET_DB_QueryStatus
+  (*lookup_instance)(void *cls,
+                     const char *id,
+                     bool active_only,
+                     TALER_MERCHANTDB_InstanceCallback cb,
+                     void *cb_cls);
+
+  /**
+   * Lookup authentication data of an instance.
+   *
+   * @param cls closure
+   * @param instance_id instance to query
+   * @param[out] ias where to store the auth data
+   */
+  enum GNUNET_DB_QueryStatus
+  (*lookup_instance_auth)(void *cls,
+                          const char *instance_id,
+                          struct TALER_MERCHANTDB_InstanceAuthSettings *ias);
+
+
   /**
    * Insert information about an instance into our database.
    *
@@ -768,7 +864,8 @@ struct TALER_MERCHANTDB_Plugin
   (*insert_instance)(void *cls,
                      const struct TALER_MerchantPublicKeyP *merchant_pub,
                      const struct TALER_MerchantPrivateKeyP *merchant_priv,
-                     const struct TALER_MERCHANTDB_InstanceSettings *is);
+                     const struct TALER_MERCHANTDB_InstanceSettings *is,
+                     const struct TALER_MERCHANTDB_InstanceAuthSettings *ias);
 
   /**
    * Insert information about an instance's account into our database.
@@ -819,17 +916,48 @@ struct TALER_MERCHANTDB_Plugin
   (*update_instance)(void *cls,
                      const struct TALER_MERCHANTDB_InstanceSettings *is);
 
+  /**
+   * Update information about an instance's authentication settings
+   * into our database.
+   *
+   * @param cls closure
+   * @param merchant_id merchant backend instance ID
+   * @param ias instance auth settings
+   * @return database result code
+   */
+  enum GNUNET_DB_QueryStatus
+  (*update_instance_auth)(void *cls,
+                          const char *merchant_id,
+                          const struct
+                          TALER_MERCHANTDB_InstanceAuthSettings *ias);
+
   /**
    * Set an instance's account in our database to "inactive".
    *
    * @param cls closure
+   * @param merchant_id merchant backend instance ID
    * @param h_wire hash of the wire account to set to inactive
    * @return database result code
    */
   enum GNUNET_DB_QueryStatus
   (*inactivate_account)(void *cls,
+                        const char *merchant_id,
                         const struct GNUNET_HashCode *h_wire);
 
+
+  /**
+   * Set an instance's account in our database to "active".
+   *
+   * @param cls closure
+   * @param merchant_id merchant backend instance ID
+   * @param h_wire hash of the wire account to set to active
+   * @return database result code
+   */
+  enum GNUNET_DB_QueryStatus
+  (*activate_account)(void *cls,
+                      const char *merchant_id,
+                      const struct GNUNET_HashCode *h_wire);
+
   /**
    * Lookup all of the products the given instance has configured.
    *
@@ -902,7 +1030,7 @@ struct TALER_MERCHANTDB_Plugin
    * @param product_id product to lookup
    * @param pd set to the product details on success, can be NULL
    *             (in that case we only want to check if the product exists);
-   *             total_sold in @a pd is ignored, total_lost must not
+   *             total_sold in @a pd is ignored (!), total_lost must not
    *             exceed total_stock minus the existing total_sold;
    *             total_sold and total_stock must be larger or equal to
    *             the existing value;
@@ -938,6 +1066,17 @@ struct TALER_MERCHANTDB_Plugin
                   struct GNUNET_TIME_Absolute expiration_time);
 
 
+  /**
+   * Release all expired product locks, including
+   * those from expired offers -- across all
+   * instances.
+   *
+   * @param cls closure
+   */
+  void
+  (*expire_locks)(void *cls);
+
+
   /**
    * Delete information about an order. Note that the transaction must
    * enforce that the order is not awaiting payment anymore.
@@ -1018,7 +1157,7 @@ struct TALER_MERCHANTDB_Plugin
    * @param cls closure
    * @param instance_id identifies the instance responsible for the order
    * @param order_id alphanumeric string that uniquely identifies the order
-   * @param h_post_order hash of the POST data for idempotency checks
+   * @param h_post_data hash of the POST data for idempotency checks
    * @param pay_deadline how long does the customer have to pay for the order
    * @param claim_token token to use for access control
    * @param contract_terms proposal data to store
@@ -1028,7 +1167,7 @@ struct TALER_MERCHANTDB_Plugin
   (*insert_order)(void *cls,
                   const char *instance_id,
                   const char *order_id,
-                  const struct GNUNET_HashCode *h_post_order,
+                  const struct GNUNET_HashCode *h_post_data,
                   struct GNUNET_TIME_Absolute pay_deadline,
                   const struct TALER_ClaimTokenP *claim_token,
                   const json_t *contract_terms);
@@ -1077,6 +1216,7 @@ struct TALER_MERCHANTDB_Plugin
    * @param order_id order_id used to lookup.
    * @param[out] contract_terms where to store the result, NULL to only check for existence
    * @param[out] order_serial set to the order's serial number
+   * @param[out] claim_token set to the claim token, NULL to only check for existence
    * @return transaction status
    */
   enum GNUNET_DB_QueryStatus
@@ -1084,7 +1224,8 @@ struct TALER_MERCHANTDB_Plugin
                            const char *instance_id,
                            const char *order_id,
                            json_t **contract_terms,
-                           uint64_t *order_serial);
+                           uint64_t *order_serial,
+                           struct TALER_ClaimTokenP *claim_token);
 
 
   /**
@@ -1501,6 +1642,38 @@ struct TALER_MERCHANTDB_Plugin
     bool confirmed);
 
 
+  /**
+   * Delete information about a transfer. Note that transfers
+   * confirmed by the exchange cannot be deleted anymore.
+   *
+   * @param cls closure
+   * @param instance_id instance to delete transfer of
+   * @param transfer_serial_id transfer to delete
+   * @return DB status code, #GNUNET_DB_STATUS_SUCCESS_NO_RESULTS
+   *           if deletion is prohibited OR transfer is unknown
+   */
+  enum GNUNET_DB_QueryStatus
+  (*delete_transfer)(void *cls,
+                     const char *instance_id,
+                     uint64_t transfer_serial_id);
+
+
+  /**
+   * Check if information about a transfer exists with the
+   * backend.  Returns no data, only the query status.
+   *
+   * @param cls closure
+   * @param instance_id instance to delete transfer of
+   * @param transfer_serial_id transfer to delete
+   * @return DB status code, #GNUNET_DB_STATUS_SUCCESS_ONE_RESULT
+   *           if the transfer record exists
+   */
+  enum GNUNET_DB_QueryStatus
+  (*check_transfer_exists)(void *cls,
+                           const char *instance_id,
+                           uint64_t transfer_serial_id);
+
+
   /**
    * Lookup account serial by payto URI.
    *
@@ -1526,7 +1699,9 @@ struct TALER_MERCHANTDB_Plugin
    * @param payto_uri what is the merchant's bank account that received the transfer
    * @param wtid identifier of the wire transfer
    * @param td transfer details to store
-   * @return transaction status
+   * @return transaction status,
+   *   #GNUNET_DB_STATUS_SUCCESS_NO_RESULTS if the @a wtid and @a exchange_uri are not known for this @a instance_id
+   *   #GNUNET_DB_STATUS_SUCCESS_ONE_RESULT on success
    */
   enum GNUNET_DB_QueryStatus
   (*insert_transfer_details)(
@@ -1592,25 +1767,31 @@ struct TALER_MERCHANTDB_Plugin
    * Lookup transfer status.
    *
    * @param cls closure
+   * @param instance_id the instance to look up details at
    * @param exchange_url the exchange that made the transfer
    * @param payto_uri account that received the transfer
    * @param wtid wire transfer subject
    * @param[out] total_amount amount that was debited from our
    *             aggregate balance at the exchange (in total, sum of
    *             the wire transfer amount and the @a wire_fee)
-   * @param[out] wire_fee the wire fee the exchange charged
-   * @param[out] execution_time when the transfer was executed by the exchange
+   * @param[out] wire_fee the wire fee the exchange charged (only set if @a have_exchange_sig is true)
+   * @param[out] exchange_amount the amount the exchange claims was transferred (only set if @a have_exchange_sig is true)
+   * @param[out] execution_time when the transfer was executed by the exchange (only set if @a have_exchange_sig is true)
+   * @param[out] have_exchange_sig do we have a response from the exchange about this transfer
    * @param[out] verified did we confirm the transfer was OK
    * @return transaction status
    */
   enum GNUNET_DB_QueryStatus
   (*lookup_transfer)(
     void *cls,
+    const char *instance_id,
     const char *exchange_url,
     const struct TALER_WireTransferIdentifierRawP *wtid,
     struct TALER_Amount *total_amount,
     struct TALER_Amount *wire_fee,
+    struct TALER_Amount *exchange_amount,
     struct GNUNET_TIME_Absolute *execution_time,
+    bool *have_exchange_sig,
     bool *verified);
 
 
@@ -1741,6 +1922,7 @@ struct TALER_MERCHANTDB_Plugin
    * @param reserve_priv which reserve is topped up or created
    * @param reserve_pub which reserve is topped up or created
    * @param exchange_url what URL is the exchange reachable at where the reserve is located
+   * @param payto_uri URI to fund the reserve
    * @param initial_balance how much money will be added to the reserve
    * @param expiration when does the reserve expire?
    * @return transaction status, usually
@@ -1752,6 +1934,7 @@ struct TALER_MERCHANTDB_Plugin
                     const struct TALER_ReservePrivateKeyP *reserve_priv,
                     const struct TALER_ReservePublicKeyP *reserve_pub,
                     const char *exchange_url,
+                    const char *payto_uri,
                     const struct TALER_Amount *initial_balance,
                     struct GNUNET_TIME_Absolute expiration);
 
@@ -1874,14 +2057,13 @@ struct TALER_MERCHANTDB_Plugin
    * @param[out] expiration set to when the tip expires
    * @return transaction status,
    *      #TALER_EC_MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_EXPIRED if the reserve is known but has expired
-   *      #TALER_EC_TIP_AUTHORIZE_RESERVE_NOT_FOUND if the reserve is not known
+   *      #TALER_EC_MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_NOT_FOUND if the reserve is not known
    *      #TALER_EC_MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_INSUFFICIENT_FUNDS if the reserve has insufficient funds left
-   *      #TALER_EC_GENERIC_DB_START_FAILED__TIP_AUTHORIZE_DB_START_FAILURE on hard DB errors
-   *      #TALER_EC_GENERIC_DB_FETCH_FAILED__TIP_AUTHORIZE_DB_LOOKUP_RESERVE_FAILURE on hard DB errors
-   *      #TALER_EC_GENERIC_DB_STORE_FAILED__TIP_AUTHORIZE_DB_UPDATE_RESERVE_FAILURE on hard DB errors
-   *      #TALER_EC_GENERIC_DB_INVARIANT_FAILURE__TIP_AUTHORIZE_DB_RESERVE_INVARIANT_FAILURE on hard DB errors
-   *      #TALER_EC_GENERIC_DB_START_FAILED__TIP_AUTHORIZE_DB_START_FAILURE on hard DB errors
-   *      #TALER_EC_GENERIC_DB_SOFT_FAILURE__TIP_AUTHORIZE_DB_SERIALIZATION_FAILURE on soft DB errors (client should retry)
+   *      #TALER_EC_GENERIC_DB_START_FAILED on hard DB errors
+   *      #TALER_EC_GENERIC_DB_FETCH_FAILED on hard DB errors
+   *      #TALER_EC_GENERIC_DB_STORE_FAILED on hard DB errors
+   *      #TALER_EC_GENERIC_DB_INVARIANT_FAILURE on hard DB errors
+   *      #TALER_EC_GENERIC_DB_SOFT_FAILURE on soft DB errors (client should retry)
    *      #TALER_EC_NONE upon success
    */
   enum TALER_ErrorCode
