@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2020 Taler Systems SA
+  Copyright (C) 2020-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
@@ -173,18 +173,13 @@ TALER_MERCHANT_order_paid (
   struct TALER_MERCHANT_OrderPaidHandle *oph;
   json_t *req_obj;
 
-  req_obj = json_pack ("{s:o, s:o, s:s}",
-                       "sig",
-                       GNUNET_JSON_from_data_auto (merchant_sig),
-                       "h_contract",
-                       GNUNET_JSON_from_data_auto (h_contract_terms),
-                       "session_id",
-                       session_id);
-  if (NULL == req_obj)
-  {
-    GNUNET_break (0);
-    return NULL;
-  }
+  req_obj = GNUNET_JSON_PACK (
+    GNUNET_JSON_pack_data_auto ("sig",
+                                merchant_sig),
+    GNUNET_JSON_pack_data_auto ("h_contract",
+                                h_contract_terms),
+    GNUNET_JSON_pack_string ("session_id",
+                             session_id));
   oph = GNUNET_new (struct TALER_MERCHANT_OrderPaidHandle);
   oph->ctx = ctx;
   oph->paid_cb = paid_cb;
