@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015, 2016, 2017, 2020 Taler Systems SA
+  Copyright (C) 2014-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
@@ -264,7 +264,7 @@ handle_pay_finished (void *cls,
     if (oph->am_wallet)
     {
       /* Here we can (and should) verify the merchant's signature */
-      struct PaymentResponsePS pr = {
+      struct TALER_PaymentResponsePS pr = {
         .purpose.purpose = htonl (TALER_SIGNATURE_MERCHANT_PAYMENT_OK),
         .purpose.size = htonl (sizeof (pr)),
         .h_contract_terms = oph->h_contract_terms
@@ -367,19 +367,19 @@ handle_pay_finished (void *cls,
        denomination key of a coin involved has expired.
        Might be a disagreement in timestamps? Still, pass on to application. */
     break;
-  case MHD_HTTP_FAILED_DEPENDENCY:
-    TALER_MERCHANT_parse_error_details_ (json,
-                                         response_code,
-                                         &hr);
-    /* Nothing really to verify, the merchant is blaming the exchange.
-       We should pass the JSON reply to the application */
-    break;
   case MHD_HTTP_INTERNAL_SERVER_ERROR:
     hr.ec = TALER_JSON_get_error_code (json);
     hr.hint = TALER_JSON_get_error_hint (json);
     /* Server had an internal issue; we should retry,
        but this API leaves this to the application */
     break;
+  case MHD_HTTP_BAD_GATEWAY:
+    /* Nothing really to verify, the merchant is blaming the exchange.
+       We should pass the JSON reply to the application */
+    TALER_MERCHANT_parse_error_details_ (json,
+                                         response_code,
+                                         &hr);
+    break;
   case MHD_HTTP_SERVICE_UNAVAILABLE:
     TALER_MERCHANT_parse_error_details_ (json,
                                          response_code,
@@ -387,6 +387,13 @@ handle_pay_finished (void *cls,
     /* Exchange couldn't respond properly; the retry is
        left to the application */
     break;
+  case MHD_HTTP_GATEWAY_TIMEOUT:
+    TALER_MERCHANT_parse_error_details_ (json,
+                                         response_code,
+                                         &hr);
+    /* Exchange couldn't respond in a timely fashion;
+       the retry is left to the application */
+    break;
   default:
     TALER_MERCHANT_parse_error_details_ (json,
                                          response_code,
@@ -474,22 +481,19 @@ TALER_MERCHANT_order_pay_frontend (
     GNUNET_CRYPTO_rsa_public_key_hash (pc->denom_pub.rsa_public_key,
                                        &denom_hash);
     /* create JSON for this coin */
-    j_coin = json_pack (
-      "{s:o, s:o,"                   /* contribution/coin_pub */
-      " s:s, s:o,"          /* exchange_url / denom_pub */
-      " s:o, s:o}",          /* ub_sig / coin_sig */
-      "contribution",
-      TALER_JSON_from_amount (&pc->amount_with_fee),
-      "coin_pub",
-      GNUNET_JSON_from_data_auto (&pc->coin_pub),
-      "exchange_url",
-      pc->exchange_url,
-      "h_denom",
-      GNUNET_JSON_from_data_auto (&denom_hash),
-      "ub_sig",
-      GNUNET_JSON_from_rsa_signature (pc->denom_sig.rsa_signature),
-      "coin_sig",
-      GNUNET_JSON_from_data_auto (&pc->coin_sig));
+    j_coin = GNUNET_JSON_PACK (
+      TALER_JSON_pack_amount ("contribution",
+                              &pc->amount_with_fee),
+      GNUNET_JSON_pack_data_auto ("coin_pub",
+                                  &pc->coin_pub),
+      GNUNET_JSON_pack_string ("exchange_url",
+                               pc->exchange_url),
+      GNUNET_JSON_pack_data_auto ("h_denom",
+                                  &denom_hash),
+      TALER_JSON_pack_denomination_signature ("ub_sig",
+                                              &pc->denom_sig),
+      GNUNET_JSON_pack_data_auto ("coin_sig",
+                                  &pc->coin_sig));
     if (0 !=
         json_array_append_new (j_coins,
                                j_coin))
@@ -500,25 +504,12 @@ TALER_MERCHANT_order_pay_frontend (
     }
   }
 
-  pay_obj = json_pack ("{ s:o }",
-                       "coins",
-                       j_coins);
-  if (NULL == pay_obj)
-  {
-    GNUNET_break (0);
-    return NULL;
-  }
-  if (NULL != session_id)
-  {
-    if (0 != json_object_set (pay_obj,
-                              "session_id",
-                              json_string (session_id)))
-    {
-      GNUNET_break (0);
-      json_decref (pay_obj);
-      return NULL;
-    }
-  }
+  pay_obj = GNUNET_JSON_PACK (
+    GNUNET_JSON_pack_array_steal ("coins",
+                                  j_coins),
+    GNUNET_JSON_pack_allow_null (
+      GNUNET_JSON_pack_string ("session_id",
+                               session_id)));
 
   oph = GNUNET_new (struct TALER_MERCHANT_OrderPayHandle);
   oph->ctx = ctx;
