@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014, 2015 Taler Systems SA
+  Copyright (C) 2014, 2015, 2020 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
@@ -64,9 +64,19 @@ run (void *cls,
   }
   if (reset_db)
   {
-    (void) plugin->drop_tables (plugin->cls);
-    TALER_MERCHANTDB_plugin_unload (plugin);
-    plugin = TALER_MERCHANTDB_plugin_load (cfg);
+    if (GNUNET_OK !=
+        plugin->drop_tables (plugin->cls))
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+                  "Failed to reset the database\n");
+    }
+  }
+  if (GNUNET_OK !=
+      plugin->create_tables (plugin->cls))
+  {
+    global_ret = 1;
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Failed to initialize tables\n");
   }
   TALER_MERCHANTDB_plugin_unload (plugin);
   GNUNET_CONFIGURATION_destroy (cfg);
@@ -86,7 +96,6 @@ main (int argc,
       char *const *argv)
 {
   struct GNUNET_GETOPT_CommandLineOption options[] = {
-
     GNUNET_GETOPT_option_flag ('r',
                                "reset",
                                "reset database (DANGEROUS: all existing data is lost!)",
@@ -94,24 +103,29 @@ main (int argc,
 
     GNUNET_GETOPT_OPTION_END
   };
+  enum GNUNET_GenericReturnValue ret;
 
   /* force linker to link against libtalerutil; if we do
      not do this, the linker may "optimize" libtalerutil
      away and skip #TALER_OS_init(), which we do need */
   (void) TALER_project_data_default ();
-  GNUNET_assert (GNUNET_OK ==
-                 GNUNET_log_setup ("taler-merchant-dbinit",
-                                   "INFO",
-                                   NULL));
   if (GNUNET_OK !=
-      GNUNET_PROGRAM_run (argc, argv,
-                          "taler-merchant-dbinit",
-                          "Initialize Taler merchant database",
-                          options,
-                          &run, NULL))
-    return 1;
+      GNUNET_STRINGS_get_utf8_args (argc, argv,
+                                    &argc, &argv))
+    return 4;
+  ret = GNUNET_PROGRAM_run (
+    argc, argv,
+    "taler-merchant-dbinit",
+    gettext_noop ("Initialize Taler merchant database"),
+    options,
+    &run, NULL);
+  GNUNET_free_nz ((void *) argv);
+  if (GNUNET_SYSERR == ret)
+    return 3;
+  if (GNUNET_NO == ret)
+    return 0;
   return global_ret;
 }
 
 
-/* end of taler-exchange-dbinit.c */
+/* end of taler-merchant-dbinit.c */
