@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  Copyright (C) 2014-2020 Taler Systems SA
+  Copyright (C) 2014-2021 Taler Systems SA
 
   TALER is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as
@@ -83,6 +83,11 @@ static struct TALER_TESTING_ExchangeConfiguration ec;
  */
 static char *merchant_url;
 
+/**
+ * Merchant instance "i1a" base URL.
+ */
+static char *merchant_url_i1a;
+
 /**
  * Merchant process.
  */
@@ -113,8 +118,8 @@ static struct GNUNET_OS_Process *merchantd;
  */
 const char *payto_uris[] = {
   PAYTO_I1,
-  "payto://sepa/CH9300762011623852957" /* Just for testing account
-                                          inactivation. */
+  "payto://iban/CH9300762011623852957?receiver-name=Test"
+  /* Just for testing account inactivation. */
 };
 
 const char *order_1_transfers[] = {
@@ -148,7 +153,7 @@ const char *order_1_forgets_3[] = {
  * @param label label to use for the command.
  */
 static struct TALER_TESTING_Command
-cmd_exec_wirewatch (char *label)
+cmd_exec_wirewatch (const char *label)
 {
   return TALER_TESTING_cmd_exec_wirewatch (label, CONFIG_FILE);
 }
@@ -435,12 +440,27 @@ run (void *cls,
                                               MHD_HTTP_OK,
                                               "deposit-simple",
                                               NULL),
+    TALER_TESTING_cmd_merchant_post_transfer2 ("post-transfer-bad",
+                                               merchant_url,
+                                               PAYTO_I1,
+                                               "EUR:4.98",
+                                               NULL /* random WTID */,
+                                               /* non-routable IP address,
+                                                  so we are sure to not get
+                                                  any reply */
+                                               "http://192.0.2.1/404/",
+                                               MHD_HTTP_GATEWAY_TIMEOUT),
     TALER_TESTING_cmd_merchant_get_transfers ("get-transfers-1",
                                               merchant_url,
                                               PAYTO_I1,
                                               MHD_HTTP_OK,
                                               "post-transfer-1",
+                                              "post-transfer-bad",
                                               NULL),
+    TALER_TESTING_cmd_merchant_delete_transfer ("delete-transfer-1",
+                                                merchant_url,
+                                                "post-transfer-bad",
+                                                MHD_HTTP_NO_CONTENT),
     TALER_TESTING_cmd_merchant_get_order2 ("get-order-merchant-1-2",
                                            merchant_url,
                                            "create-proposal-1",
@@ -536,8 +556,8 @@ run (void *cls,
                                               json_object (),
                                               "can",
                                               "EUR:1",
-                                              json_object (),
-                                              json_object (),
+                                              "data:image/jpeg;base64,RAWDATA",
+                                              json_array (),
                                               5,
                                               0,
                                               json_object (),
@@ -556,12 +576,12 @@ run (void *cls,
                                              "order-p3",
                                              GNUNET_TIME_UNIT_ZERO_ABS,
                                              GNUNET_TIME_UNIT_FOREVER_ABS,
-                                             true,
+                                             true, /* claim token */
                                              "EUR:5.0",
                                              "unsupported-wire-method",
                                              "product-3/2",
-                                             "",
-                                             NULL),
+                                             "", /* locks */
+                                             NULL /* duplicate_of */),
     TALER_TESTING_cmd_merchant_post_orders2 ("create-proposal-p3-pd-nx",
                                              merchant_url,
                                              MHD_HTTP_NOT_FOUND,
@@ -842,6 +862,106 @@ run (void *cls,
     TALER_TESTING_cmd_end ()
   };
 
+  struct TALER_TESTING_Command auth[] = {
+    TALER_TESTING_cmd_merchant_post_instances ("instance-create-i1a",
+                                               merchant_url,
+                                               "i1a",
+                                               PAYTO_I1,
+                                               "EUR",
+                                               MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_get_product ("get-nx-product-i1a-1",
+                                            merchant_url_i1a,
+                                            "nx-product",
+                                            MHD_HTTP_NOT_FOUND,
+                                            NULL),
+    TALER_TESTING_cmd_merchant_get_products ("get-i1a-products-empty-1",
+                                             merchant_url_i1a,
+                                             MHD_HTTP_OK,
+                                             NULL),
+    TALER_TESTING_cmd_merchant_post_instance_auth (
+      "instance-create-i1a-auth-fail",
+      merchant_url,
+      "i1a",
+      "my-secret",
+      MHD_HTTP_BAD_REQUEST),
+    TALER_TESTING_cmd_merchant_get_product ("get-nx-product-i1a-2",
+                                            merchant_url_i1a,
+                                            "nx-product",
+                                            MHD_HTTP_NOT_FOUND,
+                                            NULL),
+    TALER_TESTING_cmd_merchant_get_products ("get-i1a-products-empty-2",
+                                             merchant_url_i1a,
+                                             MHD_HTTP_OK,
+                                             NULL),
+    TALER_TESTING_cmd_merchant_post_instance_auth (
+      "instance-create-i1a-auth-ok",
+      merchant_url,
+      "i1a",
+      RFC_8959_PREFIX "my-secret",
+      MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_get_product ("get-nx-product-i1a-3",
+                                            merchant_url_i1a,
+                                            "nx-product",
+                                            MHD_HTTP_UNAUTHORIZED,
+                                            NULL),
+    TALER_TESTING_cmd_merchant_get_products ("get-i1a-products-empty-3",
+                                             merchant_url_i1a,
+                                             MHD_HTTP_UNAUTHORIZED,
+                                             NULL),
+    TALER_TESTING_cmd_set_authorization ("set-auth-valid",
+                                         "Bearer " RFC_8959_PREFIX "my-secret"),
+    TALER_TESTING_cmd_merchant_get_product ("get-nx-product-i1a-4",
+                                            merchant_url_i1a,
+                                            "nx-product",
+                                            MHD_HTTP_NOT_FOUND,
+                                            NULL),
+    TALER_TESTING_cmd_merchant_get_products ("get-i1a-products-empty-4",
+                                             merchant_url_i1a,
+                                             MHD_HTTP_OK,
+                                             NULL),
+    TALER_TESTING_cmd_merchant_post_instance_auth (
+      "instance-create-i1a-change-auth",
+      merchant_url_i1a,
+      NULL,
+      RFC_8959_PREFIX "my-other-secret",
+      MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_get_product ("get-nx-product-i1a-5",
+                                            merchant_url_i1a,
+                                            "nx-product",
+                                            MHD_HTTP_UNAUTHORIZED,
+                                            NULL),
+    TALER_TESTING_cmd_merchant_get_products ("get-i1a-products-empty-5",
+                                             merchant_url_i1a,
+                                             MHD_HTTP_UNAUTHORIZED,
+                                             NULL),
+    TALER_TESTING_cmd_merchant_purge_instance ("instance-delete-i1a-fail",
+                                               merchant_url_i1a,
+                                               NULL,
+                                               MHD_HTTP_UNAUTHORIZED),
+    TALER_TESTING_cmd_set_authorization (
+      "set-auth-valid-again",
+      "Bearer " RFC_8959_PREFIX "my-other-secret"),
+    TALER_TESTING_cmd_merchant_post_instance_auth (
+      "instance-create-i1a-auth-ok-idempotent",
+      merchant_url_i1a,
+      NULL,
+      RFC_8959_PREFIX "my-other-secret",
+      MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_post_instance_auth (
+      "instance-create-i1a-clear-auth",
+      merchant_url_i1a,
+      NULL,
+      NULL,
+      MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_set_authorization ("set-auth-none",
+                                         NULL),
+    TALER_TESTING_cmd_merchant_purge_instance ("instance-delete-i1a",
+                                               merchant_url_i1a,
+                                               NULL,
+                                               MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_end ()
+  };
+
   struct TALER_TESTING_Command tip[] = {
     TALER_TESTING_cmd_merchant_post_reserves ("create-reserve-tip-1",
                                               merchant_url,
@@ -853,12 +973,12 @@ run (void *cls,
                                                    "EUR:20.04",
                                                    &bc.exchange_auth,
                                                    payer_payto,
-                                                   "create-reserve-tip-1"),
+                                                   "create-reserve-tip-1",
+                                                   MHD_HTTP_OK),
+    /* We need to wait until the merchant re-tries fetching the
+       reserve from the exchange.  */
     cmd_exec_wirewatch ("wirewatch-3"),
-    /* Sleep so the merchant discovers that the reserve exists at the exchange
-       before trying to authorize a tip. */
-    TALER_TESTING_cmd_sleep ("wait-reserve-1",
-                             1),
+    TALER_TESTING_cmd_sleep ("tip-sleep", 3),
     TALER_TESTING_cmd_tip_authorize ("authorize-tip-1",
                                      merchant_url,
                                      EXCHANGE_URL,
@@ -868,7 +988,7 @@ run (void *cls,
     TALER_TESTING_cmd_tip_authorize_from_reserve ("authorize-tip-2",
                                                   merchant_url,
                                                   EXCHANGE_URL,
-                                                  "create-reserve-tip-1",
+                                                  "create-reserve-tip-1-exch",
                                                   MHD_HTTP_OK,
                                                   "tip 2",
                                                   "EUR:5.01"),
@@ -904,16 +1024,16 @@ run (void *cls,
     TALER_TESTING_cmd_merchant_get_reserves ("get-reserves-1",
                                              merchant_url,
                                              MHD_HTTP_OK,
-                                             "create-reserve-tip-1",
+                                             "create-reserve-tip-1-exch",
                                              NULL),
     TALER_TESTING_cmd_merchant_get_reserve ("get-reserve-1",
                                             merchant_url,
                                             MHD_HTTP_OK,
-                                            "create-reserve-tip-1"),
+                                            "create-reserve-tip-1-exch"),
     TALER_TESTING_cmd_merchant_get_reserve_with_tips ("get-reserve-2",
                                                       merchant_url,
                                                       MHD_HTTP_OK,
-                                                      "create-reserve-tip-1",
+                                                      "create-reserve-tip-1-exch",
                                                       "authorize-tip-1",
                                                       "authorize-tip-2",
                                                       NULL),
@@ -979,10 +1099,9 @@ run (void *cls,
                                                    "EUR:1.04",
                                                    &bc.exchange_auth,
                                                    payer_payto,
-                                                   "create-reserve-tip-2"),
+                                                   "create-reserve-tip-2",
+                                                   MHD_HTTP_OK),
     cmd_exec_wirewatch ("wirewatch-4"),
-    TALER_TESTING_cmd_sleep ("wait-reserve-2",
-                             1),
     TALER_TESTING_cmd_tip_authorize_from_reserve_with_ec (
       "authorize-tip-insufficient-funds",
       merchant_url,
@@ -1000,14 +1119,12 @@ run (void *cls,
                                           "fake-tip-authorization",
                                           pickup_amounts_1,
                                           TALER_EC_MERCHANT_GENERIC_TIP_ID_UNKNOWN),
-
     TALER_TESTING_cmd_merchant_get_reserves ("get-reserves-2",
                                              merchant_url,
                                              MHD_HTTP_OK,
                                              "create-reserve-tip-1",
                                              "create-reserve-tip-2",
                                              NULL),
-
     TALER_TESTING_cmd_merchant_delete_reserve ("delete-reserve-tip-1",
                                                merchant_url,
                                                "create-reserve-tip-1",
@@ -1131,6 +1248,22 @@ run (void *cls,
   };
 
   struct TALER_TESTING_Command commands[] = {
+    /* general setup */
+    TALER_TESTING_cmd_auditor_add ("add-auditor-OK",
+                                   MHD_HTTP_NO_CONTENT,
+                                   false),
+    TALER_TESTING_cmd_wire_add ("add-wire-account",
+                                "payto://x-taler-bank/localhost/2",
+                                MHD_HTTP_NO_CONTENT,
+                                false),
+    TALER_TESTING_cmd_exec_offline_sign_keys ("offline-sign-future-keys",
+                                              CONFIG_FILE),
+    TALER_TESTING_cmd_exec_offline_sign_fees ("offline-sign-fees",
+                                              CONFIG_FILE,
+                                              "EUR:0.01",
+                                              "EUR:0.01"),
+    TALER_TESTING_cmd_check_keys_pull_all_keys ("refetch /keys",
+                                                1),
     TALER_TESTING_cmd_batch ("orders-id",
                              get_private_order_id),
     TALER_TESTING_cmd_config ("config",
@@ -1140,6 +1273,12 @@ run (void *cls,
                                               merchant_url,
                                               MHD_HTTP_OK,
                                               NULL),
+    TALER_TESTING_cmd_merchant_post_instances ("instance-create-default-setup",
+                                               merchant_url,
+                                               "default",
+                                               PAYTO_I1,
+                                               "EUR",
+                                               MHD_HTTP_NO_CONTENT),
     TALER_TESTING_cmd_merchant_post_instances ("instance-create-i1",
                                                merchant_url,
                                                "i1",
@@ -1150,6 +1289,7 @@ run (void *cls,
                                               merchant_url,
                                               MHD_HTTP_OK,
                                               "instance-create-i1",
+                                              "instance-create-default-setup",
                                               NULL),
     TALER_TESTING_cmd_merchant_get_instance ("instances-get-i1",
                                              merchant_url,
@@ -1173,7 +1313,7 @@ run (void *cls,
                                                "USD:0.5",
                                                GNUNET_TIME_UNIT_MINUTES,
                                                GNUNET_TIME_UNIT_MINUTES,
-                                               MHD_HTTP_CONFLICT),
+                                               MHD_HTTP_BAD_REQUEST),
     TALER_TESTING_cmd_merchant_patch_instance ("instance-patch-i1",
                                                merchant_url,
                                                "i1",
@@ -1239,7 +1379,44 @@ run (void *cls,
                                                "i2",
                                                PAYTO_I1,
                                                "USD",
-                                               MHD_HTTP_CONFLICT),
+                                               MHD_HTTP_BAD_REQUEST),
+    TALER_TESTING_cmd_merchant_post_instances2 ("instance-create-ACL",
+                                                merchant_url,
+                                                "i-acl",
+                                                0, NULL,
+                                                "controlled instance",
+                                                json_pack ("{s:s}", "city",
+                                                           "shopcity"),
+                                                json_pack ("{s:s}", "city",
+                                                           "lawyercity"),
+                                                "EUR:0.1",
+                                                42,
+                                                "EUR:0.2",
+                                                GNUNET_TIME_UNIT_MINUTES,
+                                                GNUNET_TIME_UNIT_MINUTES,
+                                                // FIXME: change this back once
+                                                // we have a update auth test CMD
+                                                // RFC_8959_PREFIX "EXAMPLE",
+                                                NULL,
+                                                MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_patch_instance ("instance-patch-ACL",
+                                               merchant_url,
+                                               "i-acl",
+                                               1,
+                                               payto_uris,
+                                               "controlled instance",
+                                               json_pack ("{s:s}",
+                                                          "street",
+                                                          "bobstreet"),
+                                               json_pack ("{s:s}",
+                                                          "street",
+                                                          "bobjuryst"),
+                                               "EUR:0.1",
+                                               4,
+                                               "EUR:0.5",
+                                               GNUNET_TIME_UNIT_MINUTES,
+                                               GNUNET_TIME_UNIT_MINUTES,
+                                               MHD_HTTP_NO_CONTENT),
     TALER_TESTING_cmd_merchant_post_instances ("instance-create-i2",
                                                merchant_url,
                                                "i2",
@@ -1267,16 +1444,33 @@ run (void *cls,
                                              "i2",
                                              MHD_HTTP_NOT_FOUND,
                                              NULL),
-    TALER_TESTING_cmd_merchant_purge_instance ("instance-delete-i2-again",
+    TALER_TESTING_cmd_merchant_purge_instance ("instance-delete-then-purge-i2",
                                                merchant_url,
                                                "i2",
-                                               MHD_HTTP_NOT_FOUND),
-    TALER_TESTING_cmd_merchant_post_instances ("instance-create-default",
+                                               MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_purge_instance ("instance-purge-i1",
+                                               merchant_url,
+                                               "i1",
+                                               MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_delete_instance ("instance-purge-then-delete-i1",
+                                                merchant_url,
+                                                "i1",
+                                                MHD_HTTP_NOT_FOUND),
+    TALER_TESTING_cmd_merchant_purge_instance ("instance-purge-i-acl-middle",
+                                               merchant_url,
+                                               "i-acl",
+                                               MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_purge_instance ("instance-purge-default-middle",
                                                merchant_url,
                                                "default",
-                                               PAYTO_I1,
-                                               "EUR",
                                                MHD_HTTP_NO_CONTENT),
+    TALER_TESTING_cmd_merchant_post_instances (
+      "instance-create-default-after-purge",
+      merchant_url,
+      "default",
+      PAYTO_I1,
+      "EUR",
+      MHD_HTTP_NO_CONTENT),
     TALER_TESTING_cmd_merchant_get_products ("get-products-empty",
                                              merchant_url,
                                              MHD_HTTP_OK,
@@ -1322,8 +1516,8 @@ run (void *cls,
                                               json_pack ("{s:s}", "en", "text"),
                                               "kg",
                                               "EUR:1",
-                                              json_object (),
-                                              json_object (),
+                                              "data:image/jpeg;base64,RAWDATA",
+                                              json_array (),
                                               40,
                                               0,
                                               json_pack ("{s:s}",
@@ -1349,8 +1543,8 @@ run (void *cls,
                                               json_pack ("{s:s}", "en", "text"),
                                               "kg",
                                               "EUR:1",
-                                              json_object (),
-                                              json_object (),
+                                              "data:image/jpeg;base64,RAWDATA",
+                                              json_array (),
                                               40,
                                               0,
                                               json_pack ("{s:s}",
@@ -1389,14 +1583,6 @@ run (void *cls,
                                                merchant_url,
                                                "product-2",
                                                MHD_HTTP_CONFLICT),
-    TALER_TESTING_cmd_merchant_purge_instance ("instance-purge-i1",
-                                               merchant_url,
-                                               "i1",
-                                               MHD_HTTP_NO_CONTENT),
-    TALER_TESTING_cmd_merchant_delete_instance ("instance-delete-i1-again",
-                                                merchant_url,
-                                                "i1",
-                                                MHD_HTTP_NOT_FOUND),
     TALER_TESTING_cmd_batch ("pay",
                              pay),
     TALER_TESTING_cmd_batch ("double-spending",
@@ -1409,8 +1595,10 @@ run (void *cls,
                              refund),
     TALER_TESTING_cmd_batch ("tip",
                              tip),
+    TALER_TESTING_cmd_batch ("auth",
+                             auth),
     /**
-     * End the suite.  
+     * End the suite.
      */
     TALER_TESTING_cmd_end ()
   };
@@ -1446,6 +1634,9 @@ main (int argc,
       (merchant_url = TALER_TESTING_prepare_merchant (CONFIG_FILE)))
     return 77;
 
+  GNUNET_asprintf (&merchant_url_i1a,
+                   "%sinstances/i1a/",
+                   merchant_url);
   TALER_TESTING_cleanup_files (CONFIG_FILE);
 
   switch (TALER_TESTING_prepare_exchange (CONFIG_FILE,
