@@ -118,7 +118,8 @@ apply_forget (void *cls,
               json_t *parent)
 {
   int *res = cls;
-  if (GNUNET_OK !=
+
+  if (GNUNET_SYSERR ==
       TALER_JSON_contract_part_forget (parent,
                                        object_id))
     *res = GNUNET_SYSERR;
@@ -372,7 +373,7 @@ merchant_get_order_cb (
         }
         if (0 < gos->refunds_length)
           GNUNET_assert (GNUNET_OK ==
-                         TALER_amount_get_zero (
+                         TALER_amount_set_zero (
                            osr->details.paid.refund_amount.currency,
                            &refunded_total));
         for (unsigned int i = 0; i < gos->refunds_length; ++i)
@@ -465,7 +466,8 @@ merchant_get_order_cb (
                                           &pud))
         {
           GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                      "Taler pay uri is malformed\n");
+                      "Taler pay uri `%s' is malformed\n",
+                      osr->details.unpaid.taler_pay_uri);
           TALER_TESTING_interpreter_fail (gos->is);
           return;
         }
@@ -478,40 +480,78 @@ merchant_get_order_cb (
             TALER_TESTING_get_trait_order_id (order_cmd,
                                               0,
                                               &order_id))
+        {
+          TALER_MERCHANT_parse_pay_uri_free (&pud);
           TALER_TESTING_FAIL (gos->is);
+        }
 
         if (GNUNET_OK !=
             TALER_TESTING_get_trait_claim_token (order_cmd,
                                                  0,
                                                  &claim_token))
-          TALER_TESTING_FAIL (gos->is);
-
-        if ((0 != strcmp ("localhost:8080",
-                          pud.merchant_host)) ||
-            (NULL != pud.merchant_prefix_path) ||
-            (0 != strcmp (order_id,
-                          pud.order_id)) ||
-            (NULL != pud.ssid))
         {
-          GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                      "Order pay uri does not match\n");
-          TALER_TESTING_interpreter_fail (gos->is);
           TALER_MERCHANT_parse_pay_uri_free (&pud);
-          return;
+          TALER_TESTING_FAIL (gos->is);
+        }
+        {
+          char *port;
+          char *host;
+
+          if (GNUNET_OK !=
+              GNUNET_CONFIGURATION_get_value_string (gos->is->cfg,
+                                                     "merchant",
+                                                     "PORT",
+                                                     &port))
+          {
+            /* How did we get here without a configured port? */
+            GNUNET_break (0);
+            TALER_TESTING_interpreter_fail (gos->is);
+            TALER_MERCHANT_parse_pay_uri_free (&pud);
+            return;
+          }
+          GNUNET_asprintf (&host,
+                           "localhost:%s",
+                           port);
+          GNUNET_free (port);
+          if ((0 != strcmp (host,
+                            pud.merchant_host)) ||
+              (NULL != pud.merchant_prefix_path) ||
+              (0 != strcmp (order_id,
+                            pud.order_id)) ||
+              (NULL != pud.ssid))
+          {
+            GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                        "Order pay uri `%s' does not match, wanted %s/%s\n",
+                        osr->details.unpaid.taler_pay_uri,
+                        host,
+                        order_id);
+            TALER_TESTING_interpreter_fail (gos->is);
+            TALER_MERCHANT_parse_pay_uri_free (&pud);
+            GNUNET_free (host);
+            return;
+          }
+          GNUNET_free (host);
         }
         /* The claim token is not given in the pay uri if the order
            has been claimed already. */
-        if ((NULL != pud.claim_token) &&
-            ((NULL == claim_token) ||
-             (0 != GNUNET_memcmp (claim_token,
-                                  pud.claim_token))))
+        if ( (NULL != pud.claim_token) &&
+             ( (NULL == claim_token) ||
+               (0 != GNUNET_memcmp (claim_token,
+                                    pud.claim_token)) ) )
         {
           GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
-                      "Order pay uri does not match\n");
+                      "Order pay uri claim token does not match (%d/%d/%d/%d)\n",
+                      NULL == pud.claim_token,
+                      NULL == claim_token,
+                      (NULL != pud.claim_token) &&
+                      GNUNET_is_zero (pud.claim_token),
+                      (NULL != claim_token) &&
+                      GNUNET_is_zero (claim_token));
           TALER_TESTING_interpreter_fail (gos->is);
           TALER_MERCHANT_parse_pay_uri_free (&pud);
           return;
         }
+        TALER_MERCHANT_parse_pay_uri_free (&pud);
       }
       break;
     }
