@@ -286,7 +286,7 @@ order_cb (void *cls,
       const struct TALER_TESTING_Command *order_cmd;
       const struct TALER_ClaimTokenP *prev_token;
       struct TALER_ClaimTokenP zero_token = {0};
-      
+
       order_cmd = TALER_TESTING_interpreter_lookup_command (
         ps->is,
         ps->duplicate_of);
@@ -312,6 +312,9 @@ order_cb (void *cls,
       }
     }
     break;
+  case MHD_HTTP_NOT_FOUND:
+    TALER_TESTING_interpreter_next (ps->is);
+    return;
   case MHD_HTTP_GONE:
     TALER_TESTING_interpreter_next (ps->is);
     return;
@@ -385,7 +388,7 @@ orders_run (void *cls,
   }
 
   if (NULL == json_object_get (order,
-                               "order_id")) 
+                               "order_id"))
   {
     struct GNUNET_TIME_Absolute now;
     char *order_id;
@@ -674,7 +677,6 @@ make_order_json (const char *order_id,
                                          "$.dummy_array[*].item",
                                          &mark_forgettable,
                                          NULL));
-
   *order = json_dumps (contract_terms, 0);
   json_decref (contract_terms);
 }
